# Optimization of vessel connections

import numpy as np
import matplotlib.pyplot as plt
from scipy import interpolate, optimize
from mpl_toolkits.mplot3d import Axes3D
import numba as nb
import nlopt
from math import tanh
from geomdl import BSpline, utilities
from geomdl.visualization import VisMPL


@nb.jit(nopython=True)
def get_angle(V1,V2):
    return np.arccos(np.dot(-V1,V2))*(180/np.pi)


@nb.jit(nopython=True)
def get_vecs(vec):
    angles = np.zeros(vec.shape[0]-1)
    for i in range(vec.shape[0]-1):
        value = get_angle(vec[i,:],vec[i+1,:])
        angles[i] = 1/value
    return angles

def get_radius(curve):
    num = curve.sample_size
    rad = []
    for i in np.linspace(0,1,num=num):
        _,d,dd = curve.derivatives(i,2)
        d  = np.array(d)
        dd = np.array(dd)
        value = np.linalg.norm(d)**3/(np.linalg.norm(np.cross(d,dd))+np.finfo(float).eps)
        rad.append(value)
    return rad

def get_bounds(bounds,sample_pts):
    bounds_score = 0
    for i in range(sample_pts.shape[0]):
        if sample_pts[i,0] < bounds[0][0]:
            bounds_score += 10e4*tanh((sample_pts[i,0]**2 - bounds[0][0]**2)**(1/2))
        elif sample_pts[i,0] > bounds[0][1]:
            bounds_score += 10e4*tanh((sample_pts[i,0]**2 - bounds[0][1]**2)**(1/2))
        if sample_pts[i,1] < bounds[1][0]:
            bounds_score += 10e4*tanh((sample_pts[i,1]**2 - bounds[1][0]**2)**(1/2))
        elif sample_pts[i,1] > bounds[1][1]:
            bounds_score += 10e4*tanh((sample_pts[i,1]**2 - bounds[1][1]**2)**(1/2))
        if sample_pts[i,2] < bounds[2][0]:
            bounds_score += 10e4*tanh((sample_pts[i,2]**2 - bounds[2][0]**2)**(1/2))
        elif sample_pts[i,2] > bounds[2][1]:
            bounds_score += 10e4*tanh((sample_pts[i,2]**2 - bounds[2][1]**2)**(1/2))
    return bounds_score

@nb.jit(nopython=True)
def get_collisions(collision_vessels,R,sample_pts,radius_buffer):
    collisions = 0
    for i in range(sample_pts.shape[0]):
        dist = close_exact(collision_vessels,sample_pts[i,:],radius_buffer)
        if np.any(dist<R):
            collisions += 10e8*tanh((R - min(dist))*(1/10))
    return collisions


@nb.jit(nopython=True)
def are_collisions(collision_vessels,R,sample_pts):
    collisions = False
    for i in range(sample_pts.shape[0]):
        dist = close_exact(collision_vessels,sample_pts[i,:],radius_buffer)
        if np.any(dist<R):
            collisions = True
            break
    return collisions


@nb.jit(nopython=True)
def close_exact(data,point,radius_buffer):
    line_direction = np.zeros((data.shape[0],3))
    ss = np.zeros(data.shape[0])
    tt = np.zeros(data.shape[0])
    hh = np.zeros(data.shape[0])
    cc = np.zeros((data.shape[0],3))
    cd = np.zeros(data.shape[0])
    line_distances = np.zeros(data.shape[0])
    for i in range(data.shape[0]):
        line_direction[i,:] = (data[i,3:6] - data[i,0:3])/np.linalg.norm(data[i,3:6] - data[i,0:3])
        ss[i] = np.dot(data[i,0:3]-point,line_direction[i,:])
        tt[i] = np.dot(point-data[i,3:6],line_direction[i,:])
        d = np.array([ss[i],tt[i],0])
        hh[i] = np.max(d)
        diff = point - data[i,0:3]
        cc[i,:] = np.cross(diff,line_direction[i,:])
        cd[i] = np.linalg.norm(cc[i,:])
        line_distances[i] = np.sqrt(hh[i]**2+cd[i]**2) - data[i,6] - radius_buffer
    return line_distances

def connect_bezier(P1,P2,P3,P4,R,collision_vessels,degree=5):
    V1   = (P2 - P1)
    V1   = V1/np.linalg.norm(V1)
    V2   = (P4 - P3)
    V2   = V2/np.linalg.norm(V2)
    def create_bezier(data,V1=V1,V2=V2,R=R,P2=P2,P4=P4,degree=degree):
        CTR0 = P2 + (R+data[0])*V1
        CTR1 = P4 + (R+data[1])*V2
        CTR  = np.zeros(data.shape[0]-2+12)
        fill_CTR = np.zeros(data.shape[0]-2).reshape(-1,3)
        seg = np.linspace(0.1,0.9,num=fill_CTR.shape[0])
        for i in range(fill_CTR.shape[0]):
            fill_CTR[i,:] = CTR0*(1-seg[i]) + CTR1*(seg[i]) #V1*(1-seg[i])*R + CTR1*(seg[i]) + V2*(seg[i])*R
        fill_CTR = fill_CTR.flatten()
        CTR[0:3]     = P2
        CTR[3:6]     = CTR0
        CTR[-6:-3]   = CTR1
        CTR[-3:]     = P4
        CTR[6:-6]    += fill_CTR + data[2:]
        CTR  = CTR.reshape(-1,3).tolist()
        curve = BSpline.Curve()
        curve.degree = degree
        curve.ctrlpts = CTR
        curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
        curve.sample_size = 20
        curve.evaluate()
        return curve
    return create_bezier


def bezier_cost(data,grad,create_curve=None,R=None,P1=None,P3=None,collision_vessels=None,radius_buffer=0,bounds=None):
    curve = create_curve(data)
    pts   = np.array(curve.evalpts)
    spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
    #num = int(spline_length // R)
    #curve.sample_size = num
    #curve.evaluate()
    sample_pts = np.array(curve.evalpts)
    sample_pts_for_vec = np.vstack((P1,sample_pts,P3))
    vec = np.diff(sample_pts_for_vec,axis=0)
    if collision_vessels is not None:
        collisions = get_collisions(collision_vessels,R,sample_pts,radius_buffer)
    else:
        collisions = 0
    bound_score = get_bounds(bounds,sample_pts)
    #vec = vec/np.linalg.norm(vec,axis=0)
    #angles = get_vecs(vec)
    curve_rads = np.array(get_radius(curve))
    curve_rads = 1/(curve_rads[curve_rads<2*np.pi*R]+np.finfo(float).eps)
    a_sum = np.sum(curve_rads)
    vec = vec/np.linalg.norm(vec,axis=0)
    angles = 1/get_vecs(vec)
    if np.isclose(a_sum,0) or any(angles < 90):
        if np.any(angles < 90):
            a_sum = np.sum(1/angles)*10e8
    return bound_score+collisions #+a_sum*spline_length


def find_optimum_connection(P1,P2,P3,P4,R,collision_vessels,bounds=None,radius_buffer=0):
    create_curve = connect_bezier(P1,P2,P3,P4,R,collision_vessels)
    cost = lambda d: bezier_cost(d,None,create_curve=create_curve,R=R,P1=P1,P3=P3,collision_vessels=collision_vessels,radius_buffer=radius_buffer,bounds=bounds)
    success = False
    x0 = np.zeros(8)
    max_time = 25
    count = 1
    L = 4*R
    level_attempts = 4
    attempt = 0
    best = np.inf
    while not success:
        #print('Linking Optimization Path Search {} Level {}'.format(count,attempt))
        # add convex constraints to lower bound from boundary object
        #if bounds is None:
        lb = np.ones((2)*3+2)*(-L) + x0
        lb[0] = 0
        lb[1] = 0
        # add convex constraints to upper bound from boundary object
        ub = np.ones((2)*3+2)*(L) + x0
        #else:
        #    lb = np.ones(8)
        #    lb[0] = 0
        #    lb[1] = 0
        #    lb[2] = bounds[0][0]
        #    lb[3] = bounds[1][0]
        #    lb[4] = bounds[2][0]
        #    lb[5] = bounds[0][0]
        #    lb[6] = bounds[1][0]
        #    lb[7] = bounds[2][0]
        #    ub = np.ones(8)
        #    ub[0] = 1
        #    ub[1] = 1
        #    ub[2] = bounds[0][1]
        #    ub[3] = bounds[1][1]
        #    ub[4] = bounds[2][1]
        #    ub[5] = bounds[0][1]
        #    ub[6] = bounds[1][1]
        #    ub[7] = bounds[2][1]
        bounds = []
        for b in range(len(lb)):
            bounds.append([lb[b],ub[b]])
        #o = nlopt.opt(2,(2)*3+2)
        #o.set_min_objective(cost)
        #o.set_lower_bounds(lb)
        #o.set_upper_bounds(ub)
        #o.set_maxtime(max_time)
        #xopt = o.optimize(x0)
        res = optimize.shgo(cost,bounds=bounds,options={'maxtime':max_time,'disp':True})
        print(res)
        xopt = res.x
        curve = create_curve(xopt)
        vis_comp = VisMPL.VisCurve3D()
        curve.vis = vis_comp
        #if o.last_optimum_value() < 100 and attempt == level_attempts:
        if res.fun < 100 and attempt == level_attempts:
            L += L*0.25
            success = True
            #print(o.last_optimum_value())
        elif attempt < level_attempts:
            attempt += 1
            max_time += 5
            #if o.last_optimum_value() < best:
            if res.fun < best:
                x0 = xopt
                #best = o.last_optimum_value()
                best = res.fun
                #curve.render()
            #if np.isclose(o.last_optimum_value(),0):
            if np.isclose(res.fun,0):
                success = True
                #print('Constraints Satisfied: Exiting...')
            #curve.render()
            continue
        else:
            #print("Above link threshold {}".format(o.last_optimum_value()))
            L += L*0.25
            count += 1
            attempt = 0
            max_time = 25
            x0 = np.zeros((2)*3+2)
    curve = create_curve(xopt)
    #curve.vis  = vis_comp
    return curve


def get_optimum_link_points(P1,P2,P3,P4,R,collision_vessels,bounds=None,radius_buffer=0):
    curve = find_optimum_connection(P1,P2,P3,P4,R,collision_vessels,bounds=bounds,radius_buffer=radius_buffer)
    pts   = np.array(curve.evalpts)
    spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
    num = int(spline_length // R)+2
    curve.sample_size = num
    curve.evaluate()
    opt_pts = np.array(curve.evalpts)
    #fig = plt.figure()
    #ax  = fig.add_subplot(111,projection='3d')
    #ax.plot(opt_pts[:,0],opt_pts[:,1],opt_pts[:,2],label='final')
    #plt.show()
    return opt_pts



#!/usr/bin/env python
# coding=utf-8

import argparse
import glob
import os
import csv
import re
import sys
import scipy
import pdb
import vtk
from scipy.interpolate import interp1d

import numpy as np
from collections import defaultdict, OrderedDict

from .common import get_dict
from .get_database import input_args
from .vtk_functions import read_geo, write_geo, collect_arrays, get_all_arrays, ClosestPoints
from .get_bc_integrals import get_res_names
from .vtk_to_xdmf import write_xdmf

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n


def map_meshes(nd_id_src, nd_id_trg):
    """
    Map source mesh to target mesh
    """
    index = np.argsort(nd_id_src)
    search = np.searchsorted(nd_id_src[index], nd_id_trg)
    return index[search]


def read_hist(db, geo):
    # read from text files
    d_hist = '/home/pfaller/work/osmsc/studies/ini_1d_quad/hist/0186_0002'
    res_bc = defaultdict(dict)
    for f_name in os.listdir(d_hist):
        if 'Hist' in f_name:
            if f_name[0] == 'P':
                f = 'pressure'
            elif f_name[0] == 'Q':
                f = 'velocity'
            if 'COR' in f_name:
                t = 'coronary'
            elif 'RCR' in f_name:
                t = 'rcr'
            res_bc[f][t] = np.loadtxt(os.path.join(d_hist, f_name), skiprows=1)

    # read centerline
    cent = read_geo(db.get_centerline_path(geo)).GetOutput()
    n_point = cent.GetNumberOfPoints()
    arrays_cent, _ = get_all_arrays(cent)
    branches = arrays_cent['BranchId']

    # get time step
    nt_out = db.get_3d_increment(geo)

    # get outlets
    caps = get_caps_db(db, geo)
    del caps['inflow']
    bct = db.get_bcs(geo)['bc_type']

    # assign results to centerline
    arrays = defaultdict(lambda: defaultdict(lambda: np.zeros(n_point)))
    counter = defaultdict(int)
    for c, br in caps.items():
        # get bc type
        t = bct[c]

        # last point in branch is outlet
        i_c = np.where(branches == br)[0][-1]

        # loop result fields
        for f in res_bc.keys():
            # extract results for this outlet
            res = res_bc[f][t]
            if len(res.shape) == 2:
                res = res_bc[f][t].T[counter[t]]

            # loop all time steps and assign outlet value
            for i in np.arange(0, len(res), nt_out):
                arrays[f][f + '_' + str(i).zfill(5)][i_c] = res[i]
        counter[t] += 1

    # add to centerline
    for arr in arrays.values():
        for n, a in arr.items():
            out_array = n2v(a)
            out_array.SetName(n)
            cent.GetPointData().AddArray(out_array)

    # add empty area
    out_array = n2v(np.zeros(n_point))
    out_array.SetName('area')
    cent.GetPointData().AddArray(out_array)

    # write to file
    f_out = os.path.join(d_hist, geo + '.vtp')
    write_geo(f_out, cent)


def read_results_0d(fpath):
    """
    Read 0d simulation results from dictionary
    """
    return get_dict(fpath)


def read_results_1d(res_dir, params_file=None):
    """
    Read results from oneDSolver and store in dictionary
    Args:
        res_dir: directory containing 1D results
        params_file: optional, path to dictionary of oneDSolver input parameters

    Returns:
    Dictionary sorted as [result field][segment id][time step]
    """
    # requested output fields
    fields_res_1d = ['flow', 'pressure', 'area', 'wss', 'Re']

    # read 1D simulation results
    results_1d = {}
    for field in fields_res_1d:
        # list all output files for field
        result_list_1d = glob.glob(os.path.join(res_dir, '*branch*seg*_' + field + '.dat'))

        # loop segments
        results_1d[field] = defaultdict(dict)
        for f_res in result_list_1d:
            with open(f_res) as f:
                reader = csv.reader(f, delimiter=' ')

                # loop nodes
                results_1d_f = []
                for line in reader:
                    results_1d_f.append([float(l) for l in line if l][1:])

            # store results and GroupId
            seg = int(re.findall(r'\d+', f_res)[-1])
            branch = int(re.findall(r'\d+', f_res)[-2])
            results_1d[field][branch][seg] = np.array(results_1d_f)

    # read simulation parameters and add to result dict
    results_1d['params'] = get_dict(params_file)

    return results_1d


def write_results(f_out, cent, arrays, only_last=True):
    """
    Write results to vtp file
    """
    # export last time step (= initial conditions)
    if only_last:
        for f, a in arrays[0]['point'].items():
            out_array = n2v(a)
            out_array.SetName(f)
            cent.GetPointData().AddArray(out_array)
    # export all time steps
    else:
        for t in arrays.keys():
            for f in arrays[t]['point'].keys():
                out_array = n2v(arrays[t]['point'][f])
                out_array.SetName(f + '_' + t)
                cent.GetPointData().AddArray(out_array)

    # write to file
    write_geo(f_out, cent)


def map_rom_to_centerline(rom, geo_cent, res, time, only_last=True):
    """
    Map 0d or 1d results to centerline
    """
    # assemble output dict
    rec_dd = lambda: defaultdict(rec_dd)
    arrays = rec_dd()

    # get centerline arrays
    arrays_cent, _ = get_all_arrays(geo_cent)

    # centerline points
    points = v2n(geo_cent.GetPoints().GetData())

    # pick results
    if only_last:
        name = rom + '_int_last'
        t_vec = time[rom]
    else:
        name = rom + '_int'
        t_vec = time[rom + '_all']

    # loop all result fields
    for f in res[0].keys():
        if 'path' in f:
            continue
        array_f = np.zeros((arrays_cent['Path'].shape[0], len(t_vec)))
        n_outlet = np.zeros(arrays_cent['Path'].shape[0])
        for br in res.keys():
            # get centerline path
            path_cent = arrays_cent['Path'][arrays_cent['BranchId'] == br]
            path_cent /= path_cent[-1]

            # get 0d path
            path_0d = res[br][rom + '_path']
            path_0d /= path_0d[-1]

            # linearly interpolate results along centerline
            f_cent = interp1d(path_0d, res[br][f][name].T)(path_cent).T

            # store in global array
            array_f[arrays_cent['BranchId'] == br] = f_cent

            # add upstream part of branch within junction
            if br == 0:
                continue

            # first point of branch
            ip = np.where(arrays_cent['BranchId'] == br)[0][0]

            # centerline that passes through branch (first occurence)
            cid = np.where(arrays_cent['CenterlineId'][ip])[0][0]

            # id of upstream junction
            jc = arrays_cent['BifurcationId'][ip - 1]

            # centerline within junction
            jc_cent = np.where(np.logical_and(arrays_cent['BifurcationId'] == jc,
                                              arrays_cent['CenterlineId'][:, cid]))[0]

            # length of centerline within junction
            jc_path = np.append(0, np.cumsum(np.linalg.norm(np.diff(points[jc_cent], axis=0), axis=1)))
            jc_path /= jc_path[-1]

            # results at upstream branch
            res_br_u = res[arrays_cent['BranchId'][jc_cent[0] - 1]][f][name]

            # results at beginning and end of centerline within junction
            f0 = res_br_u[-1]
            f1 = res[br][f][name][0]

            # map 1d results to centerline using paths
            array_f[jc_cent] += interp1d([0, 1], np.vstack((f0, f1)).T, fill_value='extrapolate')(jc_path).T

            # count number of outlets of this junction
            n_outlet[jc_cent] += 1

            # normalize by number of outlets
        array_f[n_outlet > 0] = (array_f[n_outlet > 0].T / n_outlet[n_outlet > 0]).T

        # assemble time steps
        if only_last:
            arrays[0]['point'][f] = array_f[:, -1]
        else:
            for i, t in enumerate(t_vec):
                arrays[str(t)]['point'][f] = array_f[:, i]

    return arrays


def load_results_3d(f_res_3d):
    """
    Read 3d results embedded in centerline and sort according to branch at time step
    """
    # read 1d geometry
    reader = read_geo(f_res_3d).GetOutput()
    res = collect_arrays(reader.GetPointData())

    # names of output arrays
    res_names = get_res_names(reader, ['pressure', 'velocity'])

    # get time steps
    has_time = np.all(['_' in k for k in res_names])

    if has_time:
        times = np.unique([float(k.split('_')[1]) for k in res_names])
    else:
        times = np.zeros(1)

    # get branch ids
    branches = np.unique(res['BranchId']).tolist()
    if -1 in branches:
        branches.remove(-1)

    # add time
    out = {'time': times}

    # initilize output arrays [time step, branch]
    for f in res_names:
        if has_time:
            name = f.split('_')[0]
        else:
            name = f
        out[name] = {}
        for br in branches:
            ids = res['BranchId'] == br
            out[name][br] = np.zeros((times.shape[0], np.sum(ids)))

    # read branch-wise results from geometry
    for f in res_names:
        if has_time:
            name, time = f.split('_')
        else:
            name, time = f, 0.0
        for br in branches:
            ids = res['BranchId'] == br
            out[name][br][float(time) == times] = res[f][ids]

    # add area (identical for all time steps)
    out['area'] = {}
    for br in branches:
        ids = res['BranchId'] == br
        out['area'][br] = np.tile(res['area'][ids], (times.shape[0], 1))

    # rename velocity to flow
    try:
        out['flow'] = out['velocity']
    except:
        raise RuntimeError('No results in file ' + f_res_3d)
    del out['velocity']

    return out


def get_time(model, res, time, dt_3d=0, nt_3d=0, ns_3d=0, t_in=0):
    if '3d_rerun' in model:
        time[model + '_all'] = res['time'] * dt_3d
    elif '3d' in model:
        time[model] = np.array([0] + res['time'].tolist())
        time[model + '_all'] = time[model]
    elif '1d' in model:
        dt = 1e-3
        time[model + '_all'] = np.arange(0, res['pressure'][0][0].shape[1] + 1)[1:] * dt
        time[model + '_all'] = np.append(0, time[model + '_all'])
    elif '0d' in model:
        time[model + '_all'] = res['time']
    else:
        raise RuntimeError('Unknown model ' + model)

    # time steps for last cycle
    if not model == '3d':
        # how many full cycles where completed?
        n_cycle = max(1, int(time[model + '_all'][-1] // t_in))
        time[model + '_n_cycle'] = n_cycle

        # first and last time step in cycle
        t_end = t_in
        t_first = t_end * (n_cycle - 1)
        t_last = t_end * n_cycle

        # tolerance (<< time step * numstep) to prevent errors due to time step round-off
        eps = 1.0e-3

        # select last cycle and shift time to start from zero
        try:
            time[model + '_last_cycle_i'] = np.logical_and(time[model + '_all'] >= t_first - eps, time[model + '_all'] <= t_last + eps)
            time[model] = time[model + '_all'][time[model + '_last_cycle_i']] - t_first
        except:
            pdb.set_trace()
        cycle_range = []
        for i in np.arange(1, n_cycle + 1):
            t_first = t_end * (i - 1)
            t_last = t_end * i
            bound0 = time[model + '_all'] >= t_first - eps
            bound1 = time[model + '_all'] <= t_last + eps
            time[model + '_i_cycle_' + str(i)] = np.logical_and(bound0, bound1)
            time[model + '_cycle_' + str(i)] = time[model + '_all'][time[model + '_i_cycle_' + str(i)]] - t_first
            cycle_range += [np.where(time[model + '_i_cycle_' + str(i)])[0]]
        time[model + '_cycles'] = np.array(cycle_range, dtype=object)
    # elif '3d_rerun' in model:
    #     time_steps = res['time'].astype(int)
    #     pdb.set_trace()



def check_consistency(r_oned, res_1d, res_3d):
    n_br_res_1d = len(res_1d['area'].keys())
    n_br_res_3d = len(res_3d['area'].keys())
    n_br_geo_1d = np.unique(v2n(r_oned.GetOutput().GetPointData().GetArray('BranchId'))).shape[0]

    if n_br_res_1d != n_br_res_3d:
        return '1d and 3d results incosistent'

    if r_oned.GetNumberOfCells() + n_br_geo_1d != r_oned.GetNumberOfPoints():
        return '1d model connectivity inconsistent'

    return None


def get_branches(arrays):
    """
    Get list of branch IDs from point arrays
    """
    branches = np.unique(arrays['BranchId']).astype(int).tolist()
    if -1 in branches:
        branches.remove(-1)
    return branches


def get_caps_db(db, geo, f_surf=None):
    """
    Get caps for OSMSC models
    """
    return get_caps(db.get_centerline_outlet_path(geo), db.get_centerline_path(geo), f_surf)


def get_caps(f_outlet, f_centerline, f_surf=None):
    """
    Map outlet names to centerline branch id
    Args:
        f_outlet: ordered list of outlet names (created during centerline extraction)
        f_centerline: centerline geometry (.vtp)

    Returns:
        dictionary {cap name: BranchId}
    """
    caps = OrderedDict()
    caps['inflow'] = 0

    # read ordered outlet names from file
    outlet_names = []
    if not os.path.exists(f_outlet):
        return None
    with open(f_outlet) as file:
        for line in file:
            outlet_names += line.splitlines()

    # read centerline
    cent = read_geo(f_centerline).GetOutput()
    if not cent.GetPointData().HasArray('BranchId'):
        raise RuntimeError('centerline branch extraction failed')
    branch_id = v2n(cent.GetPointData().GetArray('BranchId'))

    # find outlets and store outlet name and BranchId
    ids = vtk.vtkIdList()
    i_outlet = 0

    # closest surface points
    if f_surf:
        # transfer surface ids
        surf = read_geo(f_surf).GetOutput()
        cell_to_point = vtk.vtkCellDataToPointData()
        cell_to_point.SetInputData(surf)
        cell_to_point.Update()
        face_id = v2n(cell_to_point.GetOutput().GetPointData().GetArray('BC_FaceID'))
        cp = ClosestPoints(f_surf)
        br_to_bcface = OrderedDict()
        br_to_bcface[0] = face_id[cp.search([list(cent.GetPoint(0))])[0]]

    # loop all centerline points
    for i in range(1, cent.GetNumberOfPoints()):
        cent.GetPointCells(i, ids)

        # check if cap
        if ids.GetNumberOfIds() == 1:
            # this works since the points are numbered according to the order of outlets
            caps[outlet_names[i_outlet]] = branch_id[i]

            # find closest surface point
            if f_surf:
                i_point = cp.search([list(cent.GetPoint(i))])[0]
                br_to_bcface[branch_id[i]] = face_id[i_point]

            i_outlet += 1

    if f_surf:
        return caps, br_to_bcface
    else:
        return caps


def res_1d_to_path(path, res):
    path_1d = []
    int_1d = []
    for seg, res_1d_seg in sorted(res.items()):
        # 1d results are duplicate at FE-nodes at corners of segments
        if seg == 0:
            # start with first FE-node
            i_start = 0
        else:
            # skip first FE-node (equal to last FE-node of previous segment)
            i_start = 1

        # generate path for segment FEs, assuming equidistant spacing
        p0 = path[seg]
        p1 = path[seg + 1]
        path_1d += np.linspace(p0, p1, res_1d_seg.shape[0])[i_start:].tolist()
        int_1d += res_1d_seg[i_start:].tolist()

    return np.array(path_1d), np.array(int_1d)


def collect_results(model, res, time, f_res, centerline=None, dt_3d=0, nt_3d=0, ns_3d=0, t_in=0, caps=None):
    # read results
    # todo: store 1d results in vtp as well
    if '0d' in model:
        res_in = read_results_0d(f_res)
        f_geo = centerline
        if res_in['time'][0] > 0:
            print('truncating results')
            i_start = np.argmin(np.abs(res_in['time'] - t_in))

            # truncate time
            for f in res_in.keys():
                if f == 'time':
                    res_in[f] = res_in[f][i_start:] - res_in[f][i_start]
                else:
                    for br in res_in[f].keys():
                        for n in res_in[f][br].keys():
                            res_in[f][br][n] = res_in[f][br][n][i_start:]
    elif '1d' in model:
        res_in = get_dict(f_res)
        f_geo = centerline
    elif '3d' in model:
        res_in = load_results_3d(f_res)
        f_geo = f_res
    else:
        raise ValueError('Model ' + model + ' not recognized')

    # read geometry
    geo = read_geo(f_geo)

    # extract point and cell arrays from geometry
    arrays, _ = get_all_arrays(geo.GetOutput())

    # get branches
    branches = get_branches(arrays)

    # simulation time steps
    get_time(model, res_in, time, dt_3d=dt_3d, nt_3d=nt_3d, ns_3d=ns_3d, t_in=t_in)

    # loop outlets
    for br in branches:
        # 1d-path along branch (real length units)
        branch_path = arrays['Path'][arrays['BranchId'] == br]

        # loop result fields
        for f in ['flow', 'pressure', 'area']:
            if '0d' in model:
                if f == 'area':
                    res[br][f]['0d_int'] = np.zeros(res_in['flow'][br].shape)
                else:
                    res[br][f]['0d_int'] = res_in[f][br]
                    res[br]['0d_path'] = res_in['distance'][br]
            elif '1d' in model:
                res[br]['1d_path'], res[br][f]['1d_int'] = res_1d_to_path(branch_path, res_in[f][br])
                if res[br][f]['1d_int'].shape[1] + 1 == time['1d_all'].shape[0]:
                    res[br][f]['1d_int'] = np.hstack((np.zeros((res[br][f]['1d_int'].shape[0], 1)), res[br][f]['1d_int']))
            elif '3d' in model:
                res[br][model + '_path'] = branch_path
                res[br][f][model + '_int'] = res_in[f][br].T

            # copy last time step at t=0
            if model == '3d':
                res[br][f][model + '_int'] = np.tile(res[br][f][model + '_int'], (1, 2))[:,
                                             res[br][f][model + '_int'].shape[1] - 1:]

            if br == 0:
                # inlet
                i_cap = 0
            else:
                # outlet
                i_cap = -1

            # extract cap results
            res[br][f][model + '_cap'] = res[br][f][model + '_int'][i_cap, :]

    # get last cycle
    for br in res.keys():
        for f in res[br].keys():
            if 'path' not in f:
                res[br][f][model + '_all'] = res[br][f][model + '_cap']

                if model + '_last_cycle_i' in time and len(time[model + '_last_cycle_i']) > 1:
                    res[br][f][model + '_int_last'] = res[br][f][model + '_int'][:, time[model + '_last_cycle_i']]
                    res[br][f][model + '_cap_last'] = res[br][f][model + '_cap'][time[model + '_last_cycle_i']]
                elif model == '3d':
                    res[br][f][model + '_int_last'] = res[br][f][model + '_int']
                    res[br][f][model + '_cap_last'] = res[br][f][model + '_cap']


def collect_results_spatial(model, res, time, f_res, dt_3d=0, t_in=0):
    geo = read_geo(f_res).GetOutput()

    # fields to export
    fields = ['pressure', 'velocity']

    # get all result array names
    res_names = get_res_names(geo, fields)

    # extract all point arrays
    arrays, _ = get_all_arrays(geo)

    # sort results according to GlobalNodeID
    mask = map_meshes(arrays['GlobalNodeID'], np.arange(1, geo.GetNumberOfPoints() + 1))

    # get time steps
    times = np.unique([float(k.split('_')[1]) for k in res_names])

    # simulation time steps
    get_time(model, {'time': times}, time, dt_3d, t_in)

    # initialize results
    res[model]['pressure'] = np.zeros((times.shape[0], geo.GetNumberOfPoints()))
    res[model]['velocity'] = np.zeros((times.shape[0], geo.GetNumberOfPoints(), 3))

    # extract results
    for f in res_names:
        n, t = f.split('_')
        res[model][n][float(t) == times] = arrays[f][mask]

    # extract periodic cycle
    # if model + '_last_cycle_i' in time:
    #     for n in fields:
    #         res[model][n] = res[model][n][time[model + '_last_cycle_i']]


def collect_results_db_0d(db, geo):
    f_res_0d = db.get_0d_flow_path(geo)
    f_oned = db.get_1d_geo(geo)

    if not os.path.exists(f_res_0d):
        return None, None

    time_inflow, _ = db.get_inflow_smooth(geo)

    res = defaultdict(lambda: defaultdict(dict))
    time = {}
    collect_results('0d', res, time, f_res_0d, centerline=f_oned, t_in=time_inflow[-1])

    return res, time


def collect_results_db_3d(db, geo, m):
    # initialzie results dict
    res = defaultdict(lambda: defaultdict(dict))
    time = {}

    if m == '3d':
        # get paths
        f_res_3d_osmsc = db.get_3d_flow(geo)
        if not os.path.exists(f_res_3d_osmsc):
            return None, None

        # collect osmsc results
        collect_results('3d', res, time, f_res_3d_osmsc)
    elif m == '3d_rerun':
        f_res_3d_rerun = db.get_3d_flow_rerun(geo)
        if not os.path.exists(f_res_3d_rerun):
            return None, None

        time_inflow, _ = db.get_inflow_smooth(geo)
        if time_inflow is None:
            return None, None

        # collect rerun results
        collect_results('3d_rerun', res, time, f_res_3d_rerun, dt_3d=db.get_3d_timestep(geo),
                        nt_3d=db.get_3d_increment(geo), ns_3d=db.get_3d_numstep(geo), t_in=time_inflow[-1])

    return res, time


def collect_results_db(db, geo, models, deformable=False):
    # initialzie results dict
    res = defaultdict(lambda: defaultdict(dict))
    time = {}

    # get paths
    f_res_0d = db.get_0d_flow_path(geo)
    f_res_1d = db.get_1d_flow_path(geo)
    f_res_3d = db.get_3d_flow(geo)
    f_oned = db.get_1d_geo(geo)
    f_cent = db.get_centerline_path(geo)

    # get paths for 3d models
    if deformable:
        f_res_3d_rerun = ['/home/pfaller/work/osmsc/studies/deformable/3d_flow/' + geo + '.vtp']
    else:
        f_res_3d_rerun = ['/home/pfaller/work/osmsc/studies/ini_1d_quad/3d_flow/' + geo + '.vtp',
                          '/home/pfaller/work/osmsc/studies/ini_zero/3d_flow/' + geo + '.vtp']

    time_inflow, _ = db.get_inflow_smooth(geo)

    # collect results
    if '3d_rerun' in models:
        for f_rerun in f_res_3d_rerun:
            if os.path.exists(f_rerun):
                collect_results('3d_rerun', res, time, f_rerun, t_in=time_inflow[-1], dt_3d=db.get_3d_timestep(geo), ns_3d=db.get_3d_numstep(geo))
                break
    if '3d' in models and os.path.exists(f_res_3d):
        collect_results('3d', res, time, f_res_3d)
    if '1d' in models and os.path.exists(f_res_1d):
        collect_results('1d', res, time, f_res_1d, f_oned, t_in=time_inflow[-1])
    if '0d' in models and os.path.exists(f_res_0d):
        collect_results('0d', res, time, f_res_0d, centerline=f_cent, t_in=time_inflow[-1])

    return res, time


def collect_results_db_3d_3d(db, geo):
    # initialzie results dict
    res = defaultdict(lambda: defaultdict(dict))
    time = {}

    # get paths
    f_res_3d_osmsc = db.get_3d_flow(geo)
    if not os.path.exists(f_res_3d_osmsc):
        return None, None

    # collect osmsc results
    collect_results('3d', res, time, f_res_3d_osmsc)

    f_res_3d_rerun = db.get_3d_flow_rerun(geo)
    if not os.path.exists(f_res_3d_rerun):
        return res, time

    time_inflow, _ = db.get_inflow_smooth(geo)
    if time_inflow is None:
        return res, time

    # collect rerun results
    collect_results('3d_rerun', res, time, f_res_3d_rerun, dt_3d=db.get_3d_timestep(geo),
                    nt_3d=db.get_3d_increment(geo), ns_3d=db.get_3d_numstep(geo), t_in=time_inflow[-1])

    return res, time


def collect_results_db_3d_3d_spatial(db, geo):
    # initialzie results dict
    res = defaultdict(lambda: defaultdict(dict))
    time = {}

    # get paths
    f_res_3d_osmsc = db.get_volume(geo)
    f_res_3d_rerun = db.get_res_3d_vol_rerun(geo)

    if not os.path.exists(f_res_3d_osmsc) or not os.path.exists(f_res_3d_rerun):
        return None, None

    time_inflow, _ = db.get_inflow_smooth(geo)

    if time_inflow is None:
        return None, None

    # collect results
    collect_results_spatial('3d', res, time, f_res_3d_osmsc)
    collect_results_spatial('3d_rerun', res, time, f_res_3d_rerun, dt_3d=db.get_3d_timestep(geo), t_in=time_inflow[-1])

    return res, time


def export_last(db, geo):
    f_res_1d = db.get_1d_flow_path(geo)

    time_inflow, _ = db.get_inflow_smooth(geo)

    # read results
    res = get_dict(f_res_1d)

    # get time information
    time = {}
    get_time('1d', res, time, t_in=time_inflow[-1])

    res_out = {'time': time['1d']}
    for f in res.keys():
        if f == 'params':
            continue
        res_out[f] = {}
        for br in res[f].keys():
            res_out[f][br] = {}
            for seg in res[f][br].keys():
                res_out[f][br][seg] = res[f][br][seg][:, time['1d_last_cycle_i']]
    np.save(db.gen_file('1d_flow_last', geo), res_out)


def export_rom_vtp_db(db, geo, model, only_last=True):
    # initialzie results dict
    res = defaultdict(lambda: defaultdict(dict))
    time = {}

    # get paths
    f_res_0d = db.get_0d_flow_path(geo)
    f_res_1d = db.get_1d_flow_path(geo)
    f_oned = db.get_1d_geo(geo)
    f_cent = db.get_centerline_path(geo)
    cent = read_geo(f_cent).GetOutput()

    # collect results
    time_inflow, _ = db.get_inflow_smooth(geo)
    if '1d' == model and os.path.exists(f_res_1d):
        collect_results('1d', res, time, f_res_1d, f_oned, t_in=time_inflow[-1])
        f_out = db.get_1d_flow_path_vtp(geo, only_last=only_last)
    elif '0d' == model and os.path.exists(f_res_0d):
        collect_results('0d', res, time, f_res_0d, centerline=f_cent, t_in=time_inflow[-1])
        f_out = db.get_0d_flow_path_vtp(geo, only_last=only_last)
    else:
        return

    arrays = map_rom_to_centerline(model, cent, res, time, only_last=only_last)
    pdb.set_trace()
    write_results(f_out, cent, arrays, only_last=only_last)


def main(db, geometries):
    for geo in geometries:
        print('Processing ' + geo)

        # read_hist(db, geo)
        if not os.path.exists(db.get_0d_flow_path(geo)):
            continue

        for m in ['0d', '1d']:
            export_rom_vtp_db(db, geo, m, only_last=True)
        # export_last(db, geo)


def main_cover(db, geometries):
    for geo in geometries:
        print('Processing ' + geo)

        for m in ['0d', '1d']:
            export_rom_vtp_db(db, geo, m, only_last=True)
        # export_last(db, geo)


if __name__ == '__main__':
    descr = 'Retrieve simulation results'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import numpy as np
import sys
import os
import argparse
import pdb

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt

# for styles, see https://tonysyu.github.io/raw_content/matplotlib-style-gallery/gallery.html
# plt.style.use('ggplot')

import matplotlib.image as mpimg
from matplotlib.offsetbox import TextArea, DrawingArea, OffsetImage, AnnotationBbox
import matplotlib.font_manager
from collections import OrderedDict, defaultdict
from scipy.interpolate import interp1d

from common import rec_dict, get_dict
from get_database import Database, Post, input_args
from simulation_io import get_caps_db, collect_results_db, collect_results_db_3d_3d, \
    collect_results_db_3d_3d_spatial

sys.path.append('/home/pfaller/work/repos/SimVascular_fork/Python/site-packages/')

import sv_rom_simulation as oned

plt.rcParams.update({'text.usetex': True, 'font.family': 'serif', 'font.serif': 'Computer Modern Roman'})

def add_image(db, geo, fig):
    if not os.path.exists(db.get_png(geo)):
        return
    im = plt.imread(db.get_png(geo))
    newax = fig.add_axes([-0.42, 0.1, 0.5, 0.5], anchor='E')#, zorder=-1
    newax.imshow(im)
    newax.axis('off')


def plot_1d_3d_all(db, opt, geo, res, time):
    # get post-processing constants
    post = Post()

    # get models
    models = [k[:-4] for k in time.keys() if '_all' in k]
    if '3d' in models and '3d_rerun' in models and opt['exclude_old']:
        models.remove('3d')

    # get 1d/3d map
    caps = get_caps_db(db, geo)

    fig, ax = plt.subplots(len(post.fields), len(models), figsize=(20, 10), dpi=opt['dpi'], sharex='col', sharey='row')

    for i, f in enumerate(post.fields):
        for j, m in enumerate(models):
            if len(models) > 1:
                pos = (i, j)
            else:
                pos = i
            ax[pos].grid(True)

            if opt['legend_row'] or i == 0:
                ax[pos].set_title((geo + ' ' + m + ' ' + f).replace('_', '\_'))
            if opt['legend_row'] or i == len(post.fields) - 1:
                ax[pos].set_xlabel('Time [s]')
            if opt['legend_col'] or j == 0:
                ax[pos].set_ylabel(f.capitalize() + ' [' + post.units[f] + ']')

            for c in caps.values():
                ax[pos].plot(time[m + '_all'], res[c][f][m + '_all'] * post.convert[f])

    add_image(db, geo, fig)
    fig.savefig(db.get_post_path(geo, 'all'))
    plt.close(fig)


def plot_1d_3d_cyclic(db, opt, geo, res, time):
    # get post-processing constants
    post = Post()

    # get 1d/3d map
    caps = db.get_surface_names(geo, 'caps')

    fig, ax = plt.subplots(len(post.fields), len(caps), figsize=(opt['w'], opt['h']), dpi=opt['dpi'], sharex=opt['sharex'], sharey=opt['sharey'])

    for i, f in enumerate(post.fields):
        for j, c in enumerate(caps):
            ax[i, j].grid(True)

            if opt['legend_row'] or i == 0:
                ax[i, j].set_title(c.replace('_', '\_'))
            if opt['legend_row'] or i == len(post.fields) - 1:
                ax[i, j].set_xlabel('Time [s]')
            if opt['legend_col'] or j == 0:
                ax[i, j].set_ylabel(f.capitalize() + ' [' + post.units[f] + ']')

            res_split = np.split(res[c][f]['1d_all'][:time['step_cycle'] * time['n_cycle']], time['n_cycle'])
            for r in res_split:
                ax[i, j].plot(time['1d'], r * post.convert[f], post.styles['1d'])

            ax[i, j].plot(time['3d'], res[c][f]['3d'] * post.convert[f], post.styles['3d'])

    add_image(db, geo, fig)
    fig.savefig(db.get_post_path(geo, 'cyclic'))
    plt.close(fig)


def plot_1d_3d_caps(db, opt, geo, res, time):
    # get post-processing constants
    post = Post()

    # get models
    models = [k[:-4] for k in time.keys() if '_all' in k]
    if '3d' in models and '3d_rerun' in models and opt['exclude_old']:
        models.remove('3d')

    # get 1d/3d map
    caps = get_caps_db(db, geo)
    
    # get cap names
    names = db.get_cap_names(geo)

    # get 0d bcs
    bc_0d = get_dict(db.get_bc_0D_path(geo, '3d_rerun'))
    
    if len(caps) > 50:
        dpi = opt['dpi'] // 2
        sharey = opt['sharey']
    else:
        dpi = opt['dpi']
        sharey = opt['sharey']

    fields = post.fields
    if 'area' in fields:
        fields.remove('area')

    fig, ax = plt.subplots(len(fields), len(caps), figsize=(opt['w'], opt['h']), dpi=dpi, sharex=opt['sharex'], sharey=sharey)

    for i, f in enumerate(fields):
        for j, (c, br) in enumerate(caps.items()):
            ax[i, j].grid(True)

            if opt['legend_row'] or i == 0:
                ax[i, j].set_title(names[c].replace('_', '\_'))
            if opt['legend_row'] or i == len(fields) - 1:
                ax[i, j].set_xlabel('Time [s]')
                ax[i, j].xaxis.set_tick_params(which='both', labelbottom=True)
            if opt['legend_col'] or j == 0:
                ax[i, j].set_ylabel(f.capitalize() + ' [' + post.units[f] + ']')
                ax[i, j].yaxis.set_tick_params(which='both', labelleft=True)

            lg = []
            for m in models:
                if m == '3d':
                    i0 = 1
                else:
                    i0 = 0
                ax[i, j].plot(time[m][i0:], res[br][f][m + '_cap_last'][i0:] * post.convert[f], post.styles[m], color=post.color[m])
                lg += [post.model_names[m]]

            # if f == 'pressure' and bc_0d:
            #     if br in bc_0d:
            #         ax[i, j].plot(bc_0d[br]['t'], bc_0d[br]['p'] * post.convert[f], 'k--')
            #         lg += ['0D BC']
            #     pres = res[br][f][m + '_cap']
            #     delta_p = np.abs(pres[-1] - pres[0]) / (np.max(pres) - np.min(pres))
            #     print(names[c], '{:.2e}'.format(delta_p))
            if geo == '0069_0001' and f == 'pressure':
                ax[i, j].set_ylim(55, 105)

    ax[0, 0].legend(lg, bbox_to_anchor=(-0.5, 1), loc='upper right')

    add_image(db, geo, fig)
    # fig.tight_layout(rect=[0, 0.03, 1, 0.95])
    fig.savefig(db.get_post_path(geo, 'caps'), bbox_inches='tight')
    plt.close(fig)


def plot_1d_3d_interior(db, opt, geo, res, time):
    # get post-processing constants
    post = Post()

    # get models
    models = [k[:-4] for k in time.keys() if '_all' in k]
    if '3d' in models and '3d_rerun' in models and opt['exclude_old']:
        models.remove('3d')

    # get 1d/3d map
    caps = get_caps_db(db, geo)
    cap_br = list(caps.values())
    cap_names = list(caps.keys())

    # get cap names
    names = db.get_cap_names(geo)

    if len(res) > 50:
        dpi = opt['dpi'] // 2
        sharey = opt['sharey']
    else:
        dpi = opt['dpi']
        sharey = opt['sharey']

    fig, ax = plt.subplots(len(post.fields), len(res), figsize=(opt['w'], opt['h']), dpi=dpi, sharex='col', sharey=sharey)

    # pick reference time step with highest inflow
    for m in models:
        if '3d' in m:
            m_ref = m
            t_max = {m_ref: np.argmax(res[0]['flow'][m_ref + '_cap_last'])}

    # pick closest time step for other models
    for m in models:
        if m != m_ref:
            t_max[m] = np.argmin(np.abs(time[m_ref][t_max[m_ref]] - time[m]))

    area_diff = []
    for i, f in enumerate(post.fields):
        for j, br in enumerate(res.keys()):
            if len(res.keys()) > 1:
                pos = (i, j)
            else:
                pos = i
            ax[pos].grid(True)

            if opt['legend_row'] or i == 0:
                if br in cap_br:
                    name = names[cap_names[cap_br.index(br)]]
                    if not name.isupper():
                        name = name.capitalize()
                else:
                    name = 'Branch ' + str(br)
                ax[pos].set_title(name.replace('_', '\_'))
            if opt['legend_row'] or i == len(post.fields) - 1:
                ax[pos].set_xlabel('Vessel path [-]')
                ax[pos].xaxis.set_tick_params(which='both', labelbottom=True)
            if opt['legend_col'] or j == 0:
                ax[pos].set_ylabel(f.capitalize() + ' [' + post.units[f] + ']')
                ax[pos].yaxis.set_tick_params(which='both', labelleft=True)

            lg = []
            for m in models:
                path = res[br][m + '_path']
                ax[pos].plot(path / path[-1], res[br][f][m + '_int_last'][:, t_max[m]] * post.convert[f], post.styles[m], color=post.color[m])
                lg.append(post.model_names[m])

                # maximum area change in branch over time
                if m == '1d' and f == 'area':
                    area = res[br][f][m + '_int_last'] * post.convert[f]
                    area_diff += [np.max((np.max(area, axis=1) - np.min(area, axis=1))/np.mean(area, axis=1))]

            ax[pos].set_xlim(0, 1)
            if geo == '0074_0001' and f == 'pressure':
                ax[pos].set_ylim(64, 72)
            if i == 0 and j == 0:
                ax[pos].legend(lg, bbox_to_anchor=(-0.5, 1), loc='upper right')

    add_image(db, geo, fig)
    # fig.tight_layout(rect=[0, 0.03, 1, 0.95])
    fig.savefig(db.get_post_path(geo, 'interior'), bbox_inches='tight')
    plt.close(fig)

    if '1d' in models and 'area' in post.fields:
        print('  max. area change: ' + '{:.2f}'.format(np.max(area_diff) * 100) + '%')


def plot_1d_3d_paper(db, opt, geo, res, time):
    # get post-processing constants
    post = Post()

    # get 1d/3d map
    caps = db.get_surface_names(geo, 'caps')

    # 0075_1001
    # names = {'inflow': 'Inflow', 'aorta_outflow': 'Aorta', 'Lt-carotid_segs_final': 'Left Carotid',
    #          'rt-carotid_segs_final': 'Right Carotid', 'btrunk_segs_final': 'Right Subclavian',
    #          'subclavian_segs_final': 'Left Subclavian'}
    # myorder = ['inflow', 'aorta_outflow', 'Lt-carotid_segs_final', 'rt-carotid_segs_final', 'subclavian_segs_final', 'btrunk_segs_final']
    # myorder = ['aorta_outflow', 'rt-carotid_segs_final', 'btrunk_segs_final']

    names = {'inflow': 'Aorta', 'R_int_iliac': 'Right Internal Iliac', 'R_ext_iliac': 'Right External Iliac',
             'SMA': 'Superior Mesenteric', 'IMA': 'Inferior Mesenteric'}
    myorder = ['IMA', 'R_int_iliac', 'R_ext_iliac']

    # layout
    nx = 2
    ny = 3

    # plot caps
    fig, ax = plt.subplots(ny, nx, figsize=(7, 6), dpi=opt['dpi'], sharex=True, sharey='col')
    for i, f in enumerate(post.fields):
        for j, c in enumerate(myorder):
            # pos = np.unravel_index(j, (ny, nx))
            pos = (j, i)
            ax[pos].grid(True)
            ax[pos].set_title(names[c])

            if opt['legend_row'] or pos[0] == ny - 1:
                ax[pos].set_xlabel('Time [s]')
                ax[pos].xaxis.set_tick_params(which='both', labelbottom=True)
            if opt['legend_col'] or pos[1] == 0:
                ax[pos].set_ylabel(f.capitalize() + ' [' + post.units[f] + ']')
                ax[pos].yaxis.set_tick_params(which='both', labelleft=True)

            for m in post.models:
                ax[pos].plot(time['3d'], res[c][f][m] * post.convert[f], post.styles[m], color=post.colors[m])
            plt.sca(ax[pos])
            plt.xlim([0, plt.xlim()[1]])

    fig.savefig(db.get_post_path(geo, 'paper_caps'), bbox_inches='tight', pad_inches=0.01)
    plt.close(fig)

    # pick time step
    t_max = np.argmax(res['flow']['inflow']['3d'])

    for i, f in enumerate(post.fields):
        fig, ax = plt.subplots(ny, nx, figsize=(9, 8), dpi=opt['dpi'], sharex=True, sharey=True)
        for j, c in enumerate(myorder):
            pos = np.unravel_index(j, (ny, nx))
            ax[pos].grid(True)
            ax[pos].set_title(names[c])
            ax[pos].set_ylim([0, 110])

            if opt['legend_row'] or pos[0] == ny - 1:
                ax[pos].set_xlabel('Vessel path [1]')
                ax[pos].xaxis.set_tick_params(which='both', labelbottom=True)
            if opt['legend_col'] or pos[1] == 0:
                ax[pos].set_ylabel(f.capitalize() + ' [' + post.units[f] + ']')
                ax[pos].yaxis.set_tick_params(which='both', labelleft=True)

            for m in post.models:
                ax[pos].plot(res['path'][c][m], res[c][f][m + '_int'][t_max] * post.convert[f], post.styles[m], color=post.colors[m])

        fig.savefig(db.get_post_path(geo, 'paper_interior_' + f), bbox_inches='tight', pad_inches=0)
        plt.close(fig)


def calc_error_all(db, geo, res, time):
    # reduced model to compare
    model_order = ['0d', '1d', '3d']
    for m_rom in model_order:
        if m_rom not in time:
            continue
        if m_rom == '3d' and '3d_rerun' not in time:
            continue

        # get reference solution
        models = [k[:-4] for k in time.keys() if '_all' in k]
        if '3d_rerun' in models:
            m_ref = '3d_rerun'
        else:
            # m_ref = '3d'
            # raise RuntimeError('No 3d_rerun results for ' + geo)
            return

        err = calc_error(db, geo, res, time, m_rom, m_ref)

        if m_rom == '0d':
            db.add_0d_3d_comparison(geo, err)
        elif m_rom == '1d':
            db.add_1d_3d_comparison(geo, err)
        elif m_rom == '3d':
            db.add_3d_3d_comparison(geo, err)

    if '0d' in time and '1d' in time:
        err = calc_error(db, geo, res, time, '0d', '1d')
        db.add_0d_1d_comparison(geo, err)


def calc_error(db, geo, res, time, m_rom, m_ref):
    # set post-processing fields
    fields = ['flow', 'pressure']

    # get 1d/3d map
    caps = get_caps_db(db, geo)

    # interpolate 1d to 3d in space and time (allow extrapolation due to round-off errors at bounds)
    interp = lambda x_1d, y_1d, x_3d: interp1d(x_1d, y_1d.T, fill_value='extrapolate')(x_3d)

    # relative difference between two arrays
    rel_diff = lambda a, b: np.abs((a - b) / b)

    # branches (interior and caps)
    branches = {'int': {}, 'cap': {}}
    for f in fields:
        branches['int'][f] = list(res.keys())
        branches['cap'][f] = list(caps.values())

    # skip inlet branch for flow (flow is prescribed)
    for d in branches.keys():
        if len(branches[d]['flow']) > 1:
            branches[d]['flow'].remove(0)

    # pick systole and diastole time points
    inflow = res[0]['flow'][m_ref + '_cap_last']
    times = {'sys': np.argmax(inflow), 'dia': np.argmin(inflow)}

    # get spatial errors
    err = rec_dict()
    for f in fields:
        for br in branches['int'][f]:
            # retrieve 3d results
            res_3d = res[br][f][m_ref + '_int_last']

            # map paths to interval [0, 1]
            path_1d = res[br][m_rom + '_path'] / res[br][m_rom + '_path'][-1]
            path_3d = res[br][m_ref + '_path'] / res[br][m_ref + '_path'][-1]

            # interpolate in space and time
            res_1d = interp(path_1d, res[br][f][m_rom + '_int_last'], path_3d)
            res_1d = interp(time[m_rom], res_1d, time[m_ref])

            # calculate spatial error (eliminate time dimension)
            if f == 'flow' or (f == 'area' and m_rom == '1d'):
                norm = np.max(res_3d, axis=1) - np.min(res_3d, axis=1)
            else:
                norm = np.mean(res_3d, axis=1)

            # difference between ref and rom
            diff = np.abs(res_1d - res_3d).T

            # loop through different error metrics
            for n in ['abs', 'rel']:
                if n == 'rel':
                    delta = diff / norm
                else:
                    delta = diff
                for tn, ti in times.items():
                    err[f]['spatial'][tn][n][br] = delta[ti]
                err[f]['spatial']['avg'][n][br] = np.mean(delta, axis=0)
                err[f]['spatial']['max'][n][br] = np.max(delta, axis=0)

    # get mean errors
    for f in fields:
        for m in err[f]['spatial'].keys():
            for n in err[f]['spatial'][m].keys():
                # get interior error
                for br in branches['int'][f]:
                    err[f]['int'][m][n][br] = np.mean(err[f]['spatial'][m][n][br])

                # get cap error
                for br in branches['cap'][f]:
                    if len(branches['cap'][f]) > 1 and br == 0:
                        # inlet
                        i_cap = 0
                    else:
                        # outlet
                        i_cap = -1
                    err[f]['cap'][m][n][br] = err[f]['spatial'][m][n][br][i_cap]

                # get error over all branches
                err[f]['int'][m][n]['all'] = np.mean([err[f]['int'][m][n][br] for br in branches['int'][f]])
                err[f]['cap'][m][n]['all'] = np.mean([err[f]['cap'][m][n][br] for br in branches['cap'][f]])
    return err


def main(db, geometries):
    # get post-processing constants
    post = Post()

    for geo in geometries:
        # generate plots
        print(geo)

        # read results
        deform = db.study == 'deformable'
        res, time = collect_results_db(db, geo, post.models, deformable=deform)

        # plot options
        opt = {'legend_col': False,
               'legend_row': False,
               'sharex': True,
               'sharey': 'row',
               'dpi': 200,
               'w': 0.5 * (len(db.get_surface_names(geo)) * 3 + 4),
               'h': 1 * (len(Post().fields) * 1 + 2),
               'exclude_old': True}

        # calculate error
        if ('0d' in time and '1d' in time) and '3d_rerun' in time:
            print('  claculating error')
            calc_error_all(db, geo, res, time)
    
        if '3d_rerun' in time and '1d' in time:# or '3d' in time:
            print('  plotting')
            plot_1d_3d_all(db, opt, geo, res, time)
            plot_1d_3d_caps(db, opt, geo, res, time)
            plot_1d_3d_interior(db, opt, geo, res, time)

        # plot_1d_3d_paper(db, opt, geo, res, time)
        # plot_1d_3d_cyclic(db, opt, geo, res, time)


if __name__ == '__main__':
    descr = 'Plot comparison of xd-results'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import pdb
import sys
import os
import vtk
import numpy as np
import time
import glob
import math
import argparse
from collections import defaultdict
from tqdm import tqdm
import pickle
import math

from get_database import input_args, Database, Post, SimVascular

# Read a vtp file and return the polydata
def read_polydata(filename, datatype=None):
    """
    Load the given file, and return a vtkPolyData object for it.
    Args:
        filename (str): Path to input file.
        datatype (str): Additional parameter for vtkIdList objects.
    Returns:
        polyData (vtkSTL/vtkPolyData/vtkXMLStructured/
                    vtkXMLRectilinear/vtkXMLPolydata/vtkXMLUnstructured/
                    vtkXMLImage/Tecplot): Output data.
    """

    # Check if file exists
    if not os.path.exists(filename):
        raise RuntimeError("Could not find file: %s" % filename)

    # Check filename format
    fileType = filename.split(".")[-1]
    if fileType == '':
        raise RuntimeError('The file does not have an extension')

    # Get reader
    if fileType == 'stl':
        reader = vtk.vtkSTLReader()
        reader.MergingOn()
    elif fileType == 'vtk':
        reader = vtk.vtkPolyDataReader()
    elif fileType == 'vtp':
        reader = vtk.vtkXMLPolyDataReader()
    elif fileType == 'vts':
        reader = vtk.vtkXMinkorporereLStructuredGridReader()
    elif fileType == 'vtr':
        reader = vtk.vtkXMLRectilinearGridReader()
    elif fileType == 'vtu':
        reader = vtk.vtkXMLUnstructuredGridReader()
    elif fileType == "vti":
        reader = vtk.vtkXMLImageDataReader()
    elif fileType == "np" and datatype == "vtkIdList":
        result = np.load(filename).astype(np.int)
        id_list = vtk.vtkIdList()
        id_list.SetNumberOfIds(result.shape[0])
        for i in range(result.shape[0]):
            id_list.SetId(i, result[i])
        return id_list
    else:
        raise RuntimeError('Unknown file type %s' % fileType)

    # Read
    reader.SetFileName(filename)
    reader.Update()
    polydata = reader.GetOutput()

    return polydata

def cut_plane(inp, origin, normal):
    """
    Cuts geometry at a plane
    Args:
        inp: InputConnection
        origin: cutting plane origin
        normal: cutting plane normal
    Returns:
        cut: cutter object
    """
    # define cutting plane
    plane = vtk.vtkPlane()
    plane.SetOrigin(origin[0], origin[1], origin[2])
    plane.SetNormal(normal[0], normal[1], normal[2])

    # define cutter
    cut = vtk.vtkCutter()
    cut.SetInputData(inp)
    cut.SetCutFunction(plane)
    cut.Update()
    return cut

def write_polydata(input_data, filename, datatype=None):
    """
    Write the given input data based on the file name extension.
    Args:
        input_data (vtkSTL/vtkPolyData/vtkXMLStructured/
                    vtkXMLRectilinear/vtkXMLPolydata/vtkXMLUnstructured/
                    vtkXMLImage/Tecplot): Input data.
        filename (str): Save path location.
        datatype (str): Additional parameter for vtkIdList objects.
    """
    # Check filename format
    fileType = filename.split(".")[-1]
    if fileType == '':
        raise RuntimeError('The file does not have an extension')

    # Get writer
    if fileType == 'stl':
        writer = vtk.vtkSTLWriter()
    elif fileType == 'vtk':
        writer = vtk.vtkPolyDataWriter()
    elif fileType == 'vts':
        writer = vtk.vtkXMLStructuredGridWriter()
    elif fileType == 'vtr':
        writer = vtk.vtkXMLRectilinearGridWriter()
    elif fileType == 'vtp':
        writer = vtk.vtkXMLPolyDataWriter()
    elif fileType == 'vtu':
        writer = vtk.vtkXMLUnstructuredGridWriter()
    elif fileType == "vti":
        writer = vtk.vtkXMLImageDataWriter()
    elif fileType == "np" and datatype == "vtkIdList":
        output_data = np.zeros(input_data.GetNumberOfIds())
        for i in range(input_data.GetNumberOfIds()):
            output_data[i] = input_data.GetId(i)
        output_data.dump(filename)
        return
    else:
        raise RuntimeError('Unknown file type %s' % fileType)

    # Set filename and input
    writer.SetFileName(filename)
    writer.SetInputData(input_data)
    writer.Update()

    # Write
    writer.Write()

def calcDistance2Points(model, pt1,pt2):
    if(type(pt1) is int or type(pt1) is int):
        x1,y1,z1 = model.GetPoint(pt1)
    elif(type(pt1) is list):
        x1,y1,z1 = pt1[0],pt1[1],pt1[2]
    else:
        vprint(type(pt1))
    if(type(pt2) is int or type(pt2) is int):
        x2,y2,z2 = model.GetPoint(pt2)
    else:
        x2,y2,z2 = pt2[0],pt2[1],pt2[2]
    distance = ((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)**(.5)
    return distance

class Graph:
    def __init__(self):
        self.nodes = set()
        self.edges = defaultdict(list)
        self.node_coord = dict()
        self.node_properties = dict()
        self.distances = {}

    def add_node(self, value):
        self.nodes.add(value)

    def add_node_coord(self, node, coord):
        self.node_coord[node] = coord

    def add_node_property(self,node,value):
        self.node_properties[node] = value

    def add_edge(self, from_node, to_node, distance):
        self.edges[from_node].append(to_node)
        self.edges[to_node].append(from_node)
        self.distances[(from_node, to_node)] = distance
        self.distances[(to_node, from_node)] = distance

    #adds a virtual node that collapses a group of nodes to one virtual node to make a source and sets distance to 0
    def add_virtual_node(self, v_node, zero_edge_nodes):
        self.nodes.add(v_node)
        self.node_coord[v_node] = [0,0,0]
        for i in zero_edge_nodes:
            self.edges[v_node].append(i)
            self.edges[i].append(v_node)
            self.distances[(v_node, i)] = 0
            self.distances[(i, v_node)] = 0
    
    #distance cutoff is a value that ensures RCA centerlines don't wrap onto the LV myocardium through the right ventricle
    def add_virtual_node_distances(self, v_node, edge_nodes, distances):
        self.nodes.add(v_node)
        self.node_coord[v_node] = [0,0,0]
        DISTANCE_CUTOFF = 1
        for i in edge_nodes:
            if(distances[i]<DISTANCE_CUTOFF):
                self.edges[v_node].append(i)
                self.edges[i].append(v_node)
                self.distances[(v_node, i)] = distances[i]
                self.distances[(i, v_node)] = distances[i]

    def get_node_coord(self,node):
        return self.node_coord[node]

    def get_num_of_nodes(self):
        return len(self.nodes)

#generates a graph of the mesh 
def generateGraph(mesh):
    print('Generating graph...')
    graph = Graph()
    print(mesh.GetNumberOfPoints())
    for i in tqdm(range(0,mesh.GetNumberOfPoints())):
        graph.add_node(i)
        graph.add_node_coord(i,mesh.GetPoint(i))
        connnectedPt_list = getConnectedVerticesNotIncludingSeed(mesh,i)
        for j in range(0,connnectedPt_list.GetNumberOfIds()):
            # new point to decide whether to add to patch, edge, or nothing (if already in edge)
            cpt = connnectedPt_list.GetId(j)
            graph.add_edge(i,cpt,calcDistance2Points(mesh,i,cpt))
    return graph

def getConnectedVerticesNotIncludingSeed(model, seedPt):
    cell_list = vtk.vtkIdList()
    connectedPts_list = vtk.vtkIdList()
    model.GetPointCells(seedPt,cell_list)
    for j in range(0,cell_list.GetNumberOfIds()):
        pt_list = vtk.vtkIdList()
        pt_list = model.GetCell(cell_list.GetId(j)).GetPointIds()
        for k in range(0,pt_list.GetNumberOfIds()):
            if (pt_list.GetId(k) != seedPt):
                connectedPts_list.InsertUniqueId(pt_list.GetId(k))
    return connectedPts_list

def dijsktra(graph, initial):
    visited = {}
    visited[initial] = 0
    path = {}
    path_nodes = set()

    nodes = set(graph.nodes)
    counter = 0
    pbar = tqdm(total=len(nodes))
    while nodes: 
        pbar.update(1)
        min_node = None
        for node in nodes:
            if node in visited:
                if min_node is None:
                    min_node = node
                elif visited[node] <= visited[min_node]:
                    min_node = node
        if min_node is None:
            break

        nodes.remove(min_node)
        current_weight = visited[min_node]

        for edge in graph.edges[min_node]:
            weight = current_weight + graph.distances[(min_node, edge)]
            if edge not in visited or weight <= visited[edge]:
                visited[edge] = weight
                path[edge] = min_node
                path_nodes.add(edge)
        counter += 1
    pbar.close()
    return visited, path

def dijsktra_destination(graph, initial, destinations):
    visited = {}
    visited[initial] = 0
    path = {}
    path_nodes = set()

    nodes = set(graph.nodes)
    counter = 0
    pbar = tqdm(total=len(nodes))
    while nodes: 
        pbar.update(1)
        min_node = None
        for node in nodes:
            if node in visited:
                if min_node is None:
                    min_node = node
                elif visited[node] <= visited[min_node]:
                    min_node = node
        if min_node is None:
            break

        nodes.remove(min_node)
        current_weight = visited[min_node]

        for edge in graph.edges[min_node]:
            weight = current_weight + graph.distances[(min_node, edge)]
            if edge not in visited or weight <= visited[edge]:
                visited[edge] = weight
                path[edge] = min_node
                path_nodes.add(edge)
        counter += 1
        if min_node in destinations:
            break
    pbar.close()
    return visited, path, min_node


def dijsktra_expand_properties(graph, initial, properties):
    visited = {}
    visited[initial] = 0
    path = {}
    path_nodes = set()

    nodes = set(graph.nodes)
    counter = 0
    pbar = tqdm(total=len(nodes))
    while nodes: 
        pbar.update(1)
        min_node = None
        for node in nodes:
            if node in visited:
                if min_node is None:
                    min_node = node
                elif visited[node] <= visited[min_node]:
                    min_node = node
        if min_node is None:
            break

        nodes.remove(min_node)
        current_weight = visited[min_node]

        for edge in graph.edges[min_node]:
            if min_node in properties:
                prop_dict = properties[min_node]
                diff = [0,0,0]
                vtk.vtkMath.Subtract(graph.get_node_coord(edge), graph.get_node_coord(min_node), diff)
                #multiplier = vtk.vtkMath.Dot(prop_dict['Tangent'],diff)
                #print('multiplier set to '+str(multiplier))
            multiplier = 1
            #print('multiplier set to 1')
            weight = current_weight + graph.distances[(min_node, edge)]*abs(multiplier)
            if edge not in visited or weight <= visited[edge]:
                visited[edge] = weight
                path[edge] = min_node
                path_nodes.add(edge)
                if min_node in properties:
                    properties[edge] = properties[min_node]
                    graph.add_node_property(edge,properties[min_node])
                    graph.add_node_property(min_node,properties[min_node])
        counter += 1
    pbar.close()
    return visited, path

def multipleSourceDistance(mesh,graph,v_node,child_nodes,distances,properties):
    graph.add_virtual_node(v_node,child_nodes)
    f = open("properties.pkl","wb")
    pickle.dump(properties,f)
    f.close()
    visited,path = dijsktra_expand_properties(graph,v_node,properties)
    data_array = vtk.vtkDoubleArray()
    data_array.SetName('distance_map')
    for i in range(0,mesh.GetNumberOfPoints()):
        if i in visited:
            data_array.InsertNextValue(visited[i])
        else:
            data_array.InsertNextValue(-1)
    mesh.GetPointData().AddArray(data_array)
    return mesh

#returns node ids and total distance of the shortest path between two points
def shortest_path(graph, origin, destination):
	visited, paths = dijkstra(graph, origin)
	full_path = deque()
	_destination = paths[destination]

	while _destination != origin:
		full_path.appendleft(_destination)
		_destination = paths[_destination]

	full_path.appendleft(origin)
	full_path.append(destination)

	return visited[destination], list(full_path)

# fast marching method that optimizes calculation of distances using a vtk mesh
def fastMarching(heart_graph,heart,seedPts):
	pt_set= set()
	numPts = heart.GetNumberOfPoints()
	for ptID in seedPts:
		pt_set.add(heart.FindPoint(ptID))

	#intialize edge list
	edgePt = set()
	temp_list = vtk.vtkIdList()
	pt_dist = {}
	for pt in  pt_set:
		connnectedPt_list = getConnectedVerticesNotIncludingSeed(heart,pt)
		for j in range(0,connnectedPt_list.GetNumberOfIds()):
			# new point to decide whether to add to patch, edge, or nothing (if already in edge)
			cpt = connnectedPt_list.GetId(j)
			pt_dist[cpt] = calcDistance2Points(heart,pt,cpt)
			temp_list.InsertNextId(cpt)
	for i in range(0,temp_list.GetNumberOfIds()):
		edgePt.add(temp_list.GetId(i))
		pt_set.add(temp_list.GetId(i))

	temp_list = vtk.vtkIdList()
	#search until all points are found
	while(len(edgePt) > 0):
		temp = set()
		for i in  edgePt:
			connnectedPt_list = getConnectedVerticesNotIncludingSeed(heart,i)
			for j in range(0,connnectedPt_list.GetNumberOfIds()):
				# new point to decide whether to add to patch, edge, or nothing (if already in edge)
				cpt = connnectedPt_list.GetId(j)
				if(cpt in pt_set and cpt in pt_dist):
					pt_set.add(i)
					pt_dist[i] = pt_dist[cpt] + calcDistance2Points(heart,i,cpt)
					heart_graph.add_edge(i,cpt,calcDistance2Points(heart,i,cpt))
				elif(connnectedPt_list.GetId(j) not in pt_set and cpt not in edgePt):
					temp.add(cpt)
		edgePt = temp
	data_array = np.zeros(numPts)
	for i in pt_dist:
		data_array[i] = pt_dist[i]

	vtk_array = vtk.vtkDoubleArray()
	for i in data_array:
		vtk_array.InsertNextValue(i)
	vtk_array.SetName('Point Distances')
	heart.GetPointData().AddArray(vtk_array)

	return pt_dist

def addPropertiesFromDict(mesh,dict):
    for p in range(0,mesh.GetNumberOfPoints()):
        node = mesh.GetPointData().GetArray('GlobalNodeID').GetValue(p)
        if node >= 0 and node < mesh.GetNumberOfPoints() and node in dict:
            property_dict = dict[node]
            for array_name in property_dict:
                if not mesh.GetPointData().HasArray(array_name):
                    data = vtk.vtkDoubleArray()
                    data.SetName(array_name)
                    if(type(property_dict[array_name]) is float or type(property_dict[array_name]) is int):
                        data.SetNumberOfComponents(1)
                    else:
                        data.SetNumberOfComponents(len(property_dict[array_name]))
                    data.SetNumberOfValues(mesh.GetNumberOfPoints() * len(property_dict[array_name]))
                    data.Fill(-1)
                    mesh.GetPointData().AddArray(data)
                    print(array_name + ' data array added to mesh.')
                    print(data.GetNumberOfTuples())
                if(type(property_dict[array_name]) is float or type(property_dict[array_name]) is int):
                    mesh.GetPointData().GetArray(array_name).SetValue(p,property_dict[array_name])
                else:
                    mesh.GetPointData().GetArray(array_name).SetTuple(p,property_dict[array_name])
    return mesh

def createParser():
    parser = argparse.ArgumentParser(description='Maps diameter from given centerline to the surface of a given 3D model.')
    parser.add_argument('centerline', type=str, help='the centerline to map diameters from')
    parser.add_argument('mesh', type=str, help='the mesh to map onto')
    parser.add_argument('out', type=str, help='output mesh')
    parser.add_argument('-f','-file', type=str, nargs='?', default = None, help='the pickle filename with data')
    parser.add_argument('-v', '-verbose', type=int, nargs='?', const=1, default=0, help='turn on verbosity')
    return parser

def run(args):
    mesh = read_polydata(args.mesh)
    centerline = read_polydata(args.centerline)
    cleaner = vtk.vtkCleanPolyData()
    cleaner.SetInputData(centerline)
    cleaner.PointMergingOn()
    cleaner.Update()
    centerline = cleaner.GetOutput()
    global vprint
    if args.v:
        def vprint(*args):
          # Print each argument separately so caller doesn't need to
          # stuff everything to be printed into a single string
          for arg in args:
            print(arg),
    else:
        vprint = lambda *a: None

    if(args.f==None):
        numPts = mesh.GetNumberOfPoints()
        data = [0]*numPts
        seed_pts = set()
        properties = {}
        distances = {}
        seed_pts_array = vtk.vtkDoubleArray()
        seed_pts_array.SetName('seed_pts_array')
        seed_pts_array.SetNumberOfValues(mesh.GetNumberOfPoints())
        seed_pts_array.Fill(-1)
        for j in range(0,centerline.GetPointData().GetNumberOfArrays()):
            data = vtk.vtkDoubleArray()
            data.SetName(centerline.GetPointData().GetArray(j).GetName())
            data.SetNumberOfValues(mesh.GetNumberOfPoints())
            data.Fill(-1)
            mesh.GetPointData().AddArray(data)
        #Add centerline points as seed points
        print('Adding centerline seed pts')
        for i in tqdm(range(0,centerline.GetNumberOfPoints())):
            pt = mesh.FindPoint(centerline.GetPoint(i))
            seed_pts.add(pt)
            seed_pts_array.SetValue(pt,i)
            property_dict = dict()
            for j in range(0,centerline.GetPointData().GetNumberOfArrays()):
                property_dict[str(centerline.GetPointData().GetArray(j).GetName())] = centerline.GetPointData().GetArray(j).GetTuple(i)                       
            properties[mesh.FindPoint(centerline.GetPoint(i))] = property_dict
            distances[mesh.FindPoint(centerline.GetPoint(i))] = calcDistance2Points(mesh,mesh.FindPoint(centerline.GetPoint(i)),centerline.GetPoint(i))
        #Add surface points as seed points
        # print('Adding surface seed pts')
        # for i in tqdm(range(0,surface.GetNumberOfPoints())):
        #     pt = mesh.FindPoint(surface.GetPoint(i))
        #     if(surface.GetPointData().GetArray('MaximumInscribedSphereRadius').GetValue(pt)>0):
        #         seed_pts.add(pt)
        #         seed_pts_array.SetValue(pt,i)
        #         property_dict = dict()
        #         for j in range(0,surface.GetPointData().GetNumberOfArrays()):
        #             property_dict[str(surface.GetPointData().GetArray(j).GetName())] = surface.GetPointData().GetArray(j).GetTuple(i)                       
        #         properties[mesh.FindPoint(surface.GetPoint(i))] = property_dict
        #         distances[mesh.FindPoint(surface.GetPoint(i))] = calcDistance2Points(mesh,mesh.FindPoint(surface.GetPoint(i)),surface.GetPoint(i))

        print('Found '+str(len(list(properties)))+' seed pts.')
        
        mesh.GetPointData().AddArray(seed_pts_array)
        write_polydata(mesh,os.path.basename(args.mesh).split('.')[0]+'_mapped.'+os.path.basename(args.mesh).split('.')[1])
        graph = generateGraph(mesh)
        mesh = multipleSourceDistance(mesh,graph,-1,seed_pts,distances,properties)
        f = open(os.path.splitext(args.out)[0] + ".pkl","wb")
        pickle.dump(graph.node_properties,f)
        f.close()

    mesh = read_polydata(args.mesh)
    f = open(os.path.splitext(args.out)[0] + ".pkl","rb")
    props = pickle.load(f)
    mesh = addPropertiesFromDict(mesh,props)
    write_polydata(mesh,args.out)


def main(db, geometries):
    for geo in geometries:
        # get file paths
        f_vol = os.path.join(db.get_sv_meshes(geo), geo + '.vtu')
        f_0d = db.get_0d_flow_path_vtp(geo)
        f_1d = db.get_1d_flow_path_vtp(geo)
        f_out = db.get_initial_conditions_pressure(geo)
        f_pkl = os.path.splitext(db.get_initial_conditions_pressure(geo))[0] + '.pkl'

        if os.path.exists(f_out):
            print(geo + ' done!')
            continue

        # select reduced order model
        if os.path.exists(f_1d):
            print(geo + ' using 1d')
            f_red = f_1d
        elif os.path.exists(f_0d):
            print(geo + ' using 0d')
            f_red = f_0d
        else:
            print(geo + ' no 0d/1d solution found')
            continue

        # set parameters
        args = type('', (), {})()
        args.centerline = f_red
        args.mesh = f_vol
        args.out = f_out
        args.v = 0
        if os.path.exists(f_pkl):
            args.f = f_pkl
        else:
            args.f = None

        # run projection
        run(args)


if __name__ == '__main__':
    descr = 'Check RCR boundary condition of 3d simulation'
    d, g, _ = input_args(descr)
    main(d, g)



# Competitive growth of Multiple Vascular Networks within a Shared Domain
import numpy as np
import os #remove before final
import time #not neccessary?
from ..branch_addition.check import *
from ..branch_addition.close import *
from ..branch_addition.local_func_v7 import *
from ..collision.sphere_proximity import *
from ..collision.collision import *
from ..branch_addition.add_bifurcation import *
from ..branch_addition.sample_triad import *
from ..branch_addition.triangle import * #might not need
from ..branch_addition.basis import *
#from scipy import interpolate
#from scipy.spatial import KDTree
from ..branch_addition.get_point import *
from copy import deepcopy
#from .geodesic import extract_surface,geodesic
#from scipy.sparse.csgraph import shortest_path
#from .add_geodesic_path import *
import pyvista as pv
import random
from tqdm import tqdm
from time import perf_counter

def compete_add(forest,network_ids=-1,radius_buffer=0.05,exact=True):
    #Find a viable new point to add within
    if network_ids == -1:
        active_network_ids = list(range(forest.number_of_networks))
    else:
        active_network_ids = [network_ids]
    #points,_ = forest.boundary.pick(size=prefetch,homogeneous=True,replacement=False)
    #points = points.tolist()
    #active_network_ids = deepcopy(network_ids)
    active_network_tree_ids = [list(range(forest.trees_per_network[net])) for net in active_network_ids]
    #scale = len(network_ids)
    volume = deepcopy(forest.boundary.volume)
    #forest.boundary.volume = forest.boundary.volume/scale
    #networks_copy = deepcopy(forest.networks)
    #unused_networks = list(set(all_network_ids) - set(active_network_ids))
    network_pbar = tqdm(total=len(active_network_ids),desc='Adding to networks',leave=False,position=1)
    while len(active_network_ids) > 0:
        nid = active_network_ids[0]
        tree_pbar = tqdm(total=len(active_network_tree_ids[nid]),desc='Adding to Network {} Trees'.format(nid),leave=False,position=2)
        tree_pbar.refresh()
        while len(active_network_tree_ids[nid]) > 0:
            njd = active_network_tree_ids[nid][0]
            repeat = True
            while repeat:
                ## Preallocate points to check for new vessel appending
                while len(forest.networks[nid][njd].rng_points) <= 1:
                    for idx in list(range(forest.number_of_networks)):
                        for jdx in list(range(forest.trees_per_network[idx])):
                            forest.networks[idx][jdx].rng_points = []
                    rng_points,_ = forest.boundary.pick(size=min(len(forest.boundary.tet_verts),1000),homogeneous=True,replacement=False)
                    rng_points = rng_points.tolist()
                    for n in range(len(rng_points)):
                        pt = np.array(rng_points.pop(0))
                        closest_network  = None
                        closest_distance = np.inf
                        for idx in list(range(forest.number_of_networks)):
                            for jdx in list(range(forest.trees_per_network[idx])):
                                if exact:
                                    distances = close_exact_v2(forest.networks[idx][jdx].data,pt)
                                else:
                                    _,distances = close(forest.networks[idx][jdx].data,pt)
                                if min(distances) < closest_distance:
                                    closest_network = idx
                                    closest_distance = min(distances)
                        forest.networks[closest_network][np.random.choice(len(forest.networks[closest_network]))].rng_points.insert(-1,pt.tolist())
                start = perf_counter()
                for n in range(len(forest.networks[nid][njd].rng_points)):
                    pt = np.array(forest.networks[nid][njd].rng_points.pop(0))
                    closest_network  = None
                    closest_distance = np.inf
                    for idx in list(range(forest.number_of_networks)):
                        for jdx in list(range(forest.trees_per_network[idx])):
                            if exact:
                                distances = close_exact_v2(forest.networks[idx][jdx].data,pt)
                            else:
                                _,distances = close(forest.networks[idx][jdx].data,pt)
                            if min(distances) < closest_distance:
                                closest_network = idx
                                closest_distance = min(distances)
                    forest.networks[closest_network][np.random.choice(len(forest.networks[closest_network]))].rng_points.insert(-1,pt.tolist())
                #print('End point reallocation: {}'.format(perf_counter()-start))
                ##
                #print(len(forest.networks[nid][njd].rng_points))
                #print("Check: {}".format(forest.networks[nid][njd].rng_points[0]))
                start = perf_counter()
                vessel,data,sub_division_map,sub_division_index,threshold = forest.networks[nid][njd].add(-1,0,isforest=True,radius_buffer=radius_buffer)
                #print('End vessel appending: {}'.format(perf_counter()-start))
                point = data[-2,3:6] #new terminal point
                #print(point)
                #closest_network = nid
                #_,closest_value = close_exact(forest.networks[nid][njd].data[vessel,:].reshape(1,forest.networks[nid][njd].data.shape[1]),point)
                #escape = False
                """
                for idx in list(range(forest.number_of_networks)):
                    for jdx in list(range(forest.trees_per_network[idx])):
                        if idx == nid:
                            continue
                        if exact:
                            other_vessels,distances = close_exact(forest.networks[idx][jdx].data,point)
                        else:
                            other_vessels,distances = close(forest.networks[idx][jdx].data,point)
                        minimum_distance = min(distances)
                        minimum_vessel_point = (forest.networks[idx][jdx].data[other_vessels[0],0:3] + forest.networks[idx][jdx].data[other_vessels[0],3:6])/2
                        if minimum_distance < closest_value:
                            #print('other closer')
                            vessel_within = True
                            if not forest.convex:
                                subdivisions = 5
                                for sub in range(1,2*subdivisions):
                                    tmp = point*(sub/(2*subdivisions)) + minimum_vessel_point*(1-sub/(2*subdivisions))
                                    if not forest.boundary.within(tmp[0],tmp[1],tmp[2],2):
                                        vessel_within = False
                                        #print('not within')
                                        break
                            if not vessel_within:
                                continue
                            else:
                                escape = True
                                break
                    if escape:
                        break
                """
                #if escape:
                #    #forest.networks[nid][njd] = deepcopy(networks_copy[nid][njd])
                #    continue
                start = perf_counter()
                new_vessels = np.vstack((data[vessel,:],data[-2,:],data[-1,:]))
                repeat = False
                #check_networks = unused_networks + [nid]
                check_networks = [nid]
                for nikd in check_networks:
                    if nikd == nid:
                        check_trees = list(range(len(forest.networks[nikd])))
                        check_trees.remove(njd)
                    else:
                        check_trees = list(range(len(forest.networks[nikd])))
                    for njkd in check_trees:
                        if not forest.networks[nikd][njkd].collision_free(new_vessels,radius_buffer=radius_buffer):
                            repeat = True
                            break
                    if repeat:
                        break
                #print('End collision testing: {}'.format(perf_counter()-start))
                if repeat:
                    #print("Repeating Network {} Tree {}".format(nid,njd))
                    #forest.networks[nid][njd] = deepcopy(networks_copy[nid][njd])
                    continue
                else:
                    forest.networks[nid][njd].data = data
                    forest.networks[nid][njd].parameters['edge_num'] += 2
                    forest.networks[nid][njd].sub_division_map = sub_division_map
                    forest.networks[nid][njd].sub_division_index = sub_division_index
                    #network_index = active_network_ids.index(nid)
                    #print("List {}: {},  Id: {}".format(nid,active_network_tree_ids[nid],njd))
                    active_network_tree_ids[nid].remove(njd)
                    tree_pbar.update(1)
                    _ = tree_pbar.refresh()
                    if len(active_network_tree_ids[nid]) == 0:
                        active_network_ids.remove(nid)
                        network_pbar.update(1)
                        _ = network_pbar.refresh()
                    #print("Active Networks: {}".format(active_network_ids))
                    #print("Completed Network {} Tree {}".format(nid,njd))
    forest.boundary.volume = volume



#!/usr/bin/env python

import numpy as np
import sys
import argparse
import pdb

import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from matplotlib.offsetbox import TextArea, DrawingArea, OffsetImage, AnnotationBbox
from collections import OrderedDict, defaultdict
from scipy.interpolate import interp1d

from common import rec_dict
from get_database import Database, Post, input_args
from simulation_io import get_caps_db, collect_results_db_1d_3d, collect_results_db_3d_3d, \
    collect_results_db_3d_3d_spatial


def calc_error_spatial(res, time):
    # interpolate 1d to 3d in space and time (allow extrapolation due to round-off errors at bounds)
    interp = lambda x_1d, y_1d, x_3d: interp1d(x_1d, y_1d.T, fill_value='extrapolate')(x_3d)

    fields = ['pressure', 'velocity']

    err = defaultdict(dict)
    for f in fields:
        err[f]['avg'] = []
        err[f]['max'] = []
        for i in np.arange(1, time['3d_rerun_n_cycle'] + 1):
            res_3d_osmsc = res['3d'][f]
            res_3d_rerun_time = res['3d_rerun'][f][time['3d_rerun_i_cycle_' + str(i)]]
            res_3d_rerun = interp(time['3d_rerun_cycle_' + str(i)], res_3d_rerun_time, time['3d'][1:]).T

            # calculate spatial error
            if f == 'velocity':
                delta = np.linalg.norm(res_3d_osmsc - res_3d_rerun, axis=-1)
                norm = np.mean(np.linalg.norm(res_3d_osmsc, axis=-1))
            elif f == 'pressure':
                delta = res_3d_osmsc - res_3d_rerun
                norm = np.mean(res_3d_osmsc)
            else:
                raise ValueError('Unknown field ' + f)

            diff = delta ** 2
            err[f]['avg'] += [np.sqrt(np.mean(diff)) / norm]
            err[f]['max'] += [np.sqrt(np.max(diff)) / norm]

            # calculate periodicity error
            err[f]['cyc'] += [np.abs(res_3d_rerun[-1] - res_3d_rerun[0]) / (np.max(res_3d_rerun) - np.min(res_3d_rerun))]

    for f in fields:
        for e in err[f]:
            print(f, e, err[f][e])

    # todo: write spatial error to geometry
    return err


def main(db, geometries):
    for geo in geometries:
        print('Comparing geometry ' + geo)

        # read results
        res, time = collect_results_db_3d_3d_spatial(db, geo)
        if res is None:
            continue

        err = calc_error_spatial(res, time)
        db.add_3d_3d_comparison(geo, err)


if __name__ == '__main__':
    descr = 'Plot comparison of xd-results'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import glob
import io
import os
import pdb
import re
import shutil
import sys
import time
import json

import numpy as np
from pandas import read_csv

from get_database import Database, SimVascular, Post, input_args, run_command
from get_sv_project import Project
from simulation_io import read_results_1d, get_caps_db
from common import coronary_sv_to_oned


import sv_rom_simulation as oned
import svzerodsolver as run0d


def get_params(db, geo):
    # store parameters for export
    params = {}

    if db.get_bcs(geo) is None:
        raise RuntimeError('No boundary conditions')

    n_inlet = db.count_inlets(geo)
    if n_inlet == 0:
        raise RuntimeError('3d geometry has no inlet')
    elif n_inlet > 1:
        raise RuntimeError('3d geometry has multiple inlets (' + repr(n_inlet) + ')')

    # simvascular project
    project = Project(db, geo, '')

    params['fpath_1d'] = db.get_solve_dir_1d(geo)
    fpath_geo = os.path.join(params['fpath_1d'], 'geometry')
    os.makedirs(fpath_geo, exist_ok=True)

    # reference pressure (= initial pressure?)
    if os.path.exists(db.get_bc_flow_path(geo)):
        res_3d = db.read_results(db.get_bc_flow_path(geo))
        params['pref'] = res_3d['pressure'][-1, 0]
    else:
        raise RuntimeError('3d results do not exist')

    if db.has_loop(geo):
        raise RuntimeError('3d geometry contains a loop')

    # copy surface model to folder if it exists
    if os.path.exists(db.get_surfaces(geo, 'all_exterior')):
        shutil.copy2(db.get_surfaces(geo, 'all_exterior'), fpath_geo)
    else:
        raise RuntimeError('3d geometry does not exist')

    # assume pre-computed centerlines from SimVascular C++
    params['centerlines_input_file'] = db.get_centerline_path(geo)
    if not os.path.exists(params['centerlines_input_file']):
        raise RuntimeError('centerline does not exist')

    # get boundary conditions
    params['bc_def'] = db.get_bcs(geo)
    if params['bc_def'] is None:
        raise RuntimeError('no boundary conditions')

    # write outlet boundary conditions to file if they exist
    params['bc_types'], err = project.write_bc(params['fpath_1d'], model='1d')
    if err:
        raise RuntimeError(err)

    # get inflow
    time, _ = db.get_inflow_smooth(geo)
    if time is None:
        raise RuntimeError('inflow does not exist')
    params['inflow_input_file'] = db.get_sv_flow_path(geo, '1d')

    # get outlets
    caps = get_caps_db(db, geo)
    del caps['inflow']
    params['outlets'] = list(caps.keys())

    # material parameters
    params['olufsen_k1'] = 1e7
    params['olufsen_k2'] = -20
    params['olufsen_k3'] = 1e7
    params['linear_ehr'] = 3e5
    if not deformable:
        params['mat_model'] = 'OLUFSEN'
    else:
        params['mat_model'] = 'LINEAR'

    # number of cycles to run
    params['n_cycle'] = project.estimate_n_cycle(n_default=50)

    # sub-segment size
    params['seg_min_num'] = 1
    params['seg_size'] = 999

    # FEM size
    params['min_num_elems'] = 10
    params['element_size'] = 0.01

    # mesh adaptive?
    params['seg_size_adaptive'] = True

    # set simulation time as end of 3d simulation
    params['save_data_freq'] = 1
    params['dt'] = 1.0e-3

    # number of time steps per cycle
    params['num_dts_cycle'] = int(time[-1] / params['dt'])

    # run all cycles
    params['num_dts'] = int(time[-1] * params['n_cycle'] / params['dt'] + 1.0)

    # model order (0 or 1)
    params['model_order'] = model_order
    params['solver_output_file'] = geo + '_' + str(model_order) + 'd.in'

    print('tau [cycles] = ' + str(params['n_cycle']))
    return params


def generate_1d_api(db, geo):
    import sv

    # get parameters
    params_db = get_params(db, geo)

    # create 1d simulation
    rom_simulation = sv.simulation.ROM()

    # Create 1D simulation parameters
    params = sv.simulation.ROMParameters()

    # Mesh parameters
    mesh_params = params.MeshParameters()
#    mesh_params.use_adaptive_meshing = params_db['seg_size_adaptive']
#    mesh_params.num_branch_segments = params_db['seg_min_num']
    mesh_params.element_size = params_db['element_size']

    # Model parameters
    model_params = params.ModelParameters()
    model_params.name = geo
    model_params.inlet_face_names = ['inflow']
    model_params.outlet_face_names = params_db['outlets']
    model_params.centerlines_file_name = params_db['centerlines_input_file']

    # Fluid properties
    fluid_props = params.FluidProperties()

    # Set wall properties
    if params_db['mat_model'] == 'OLUFSEN':
        material = params.WallProperties.OlufsenMaterial()
        material.k1 = params_db['olufsen_k1']
        material.k2 = params_db['olufsen_k2']
        material.k3 = params_db['olufsen_k3']
        material.exponent = 2.0
        material.pressure = params_db['pref']
    elif params_db['mat_model'] == 'LINEAR':
        material = params.WallProperties.LinearMaterial()
        material.eh_r = params_db['linear_ehr']
        material.pressure = params_db['pref']
    else:
        raise ValueError('Unknonwn material ' + params_db['mat_model'])

    # Set boundary conditions
    bcs = params.BoundaryConditions()

    # add inlet
    bcs.add_velocities(face_name='inlet', file_name=params_db['inflow_input_file'])

    # add outlets
    for cp in params_db['outlets']:
        t = params_db['bc_def']['bc_type'][cp]
        bc = params_db['bc_def']['bc'][cp]
        if 'Po' in bc and bc['Po'] != 0.0:
            raise ValueError('RCR reference pressure not implemented')
        if t == 'rcr':
            bcs.add_rcr(face_name=cp, Rp=bc['Rp'], C=bc['C'], Rd=bc['Rd'])
        elif t == 'resistance':
            bcs.add_resistance(face_name=cp, resistance=bc['R'])
        elif t == 'coronary':
            raise ValueError('Coronary BC not implemented')

    # Set solution parameters
    solution_params = params.Solution()
    solution_params.time_step = params_db['dt']
    solution_params.num_time_steps = params_db['num_dts']
    solution_params.save_frequency = params_db['save_data_freq']

    # Write a 1D solver input file
    rom_simulation.write_input_file(model=model_params,
                                    model_order=model_order,
                                    mesh=mesh_params,
                                    fluid=fluid_props,
                                    material=material,
                                    boundary_conditions=bcs,
                                    solution=solution_params,
                                    directory=params_db['fpath_1d'])
    return None


def generate_1d(db, geo):
    # get parameters
    params = get_params(db, geo)

    # simvascular project
    project = Project(db, geo, '')

    # set simulation paths
    fpath_geo = os.path.join(params['fpath_1d'], 'geometry')
    fpath_surf = os.path.join(fpath_geo, 'surfaces')
    os.makedirs(fpath_surf, exist_ok=True)

    # copy outlet surfaces
    for f in db.get_surfaces(geo, 'caps'):
        shutil.copy2(f, fpath_surf)

    # copy outlet names
    fpath_outlets = os.path.join(params['fpath_1d'], 'outlets')
    shutil.copy(db.get_centerline_outlet_path(geo), fpath_outlets)

    # write inflow
    project.write_inflow('1d')

    # try:
    if True:
        oned.run(model_order=params['model_order'],
                 boundary_surfaces_directory=fpath_surf,
                 centerlines_input_file=params['centerlines_input_file'],
                 centerlines_output_file=None,
                 compute_centerlines=False,
                 compute_mesh=True,
                 density=params['bc_def']['params']['sim_density'],
                 element_size=params['element_size'],
                 inlet_face_input_file='inflow.vtp',
                 inflow_input_file=db.get_sv_flow_path(geo, '1d'),
                 linear_material_ehr=params['linear_ehr'],
                 linear_material_pressure=params['pref'],
                 material_model=params['mat_model'],
                 mesh_output_file=geo + '.vtp',
                 min_num_elements=params['min_num_elems'],
                 model_name=geo,
                 num_time_steps=params['num_dts'],
                 olufsen_material_k1=params['olufsen_k1'],
                 olufsen_material_k2=params['olufsen_k2'],
                 olufsen_material_k3=params['olufsen_k3'],
                 olufsen_material_exp=2.0,
                 olufsen_material_pressure=params['pref'],
                 outflow_bc_input_file=params['fpath_1d'],
                 outflow_bc_type=','.join(params['bc_types']),
                 outlet_face_names_input_file=fpath_outlets,
                 output_directory=params['fpath_1d'],
                 seg_min_num=params['seg_min_num'],
                 seg_size=params['seg_size'],
                 seg_size_adaptive=params['seg_size_adaptive'],
                 solver_output_file=params['solver_output_file'],
                 save_data_frequency=params['save_data_freq'],
                 surface_model=os.path.join(fpath_geo, 'all_exterior.vtp'),
                 time_step=params['dt'],
                 uniform_bc=True,
                 units='cm',
                 viscosity=params['bc_def']['params']['sim_viscosity'],
                 wall_properties_input_file=None,
                 wall_properties_output_file=None,
                 write_mesh_file=True,
                 write_solver_file=True)
    # except Exception as e:
    #     return repr(e)

    return None

model_order = 0
zerod_cpp = True
zerod_cpp_fast = True
deformable = False

# from profilehooks import profile
# @profile
def main(db, geometries):
    # simvascular object
    sv = SimVascular()

    for geo in geometries:
        print('Running geometry ' + geo)

        # if (model_order == 0 and os.path.exists(db.get_0d_flow_path(geo))) or (model_order == 1 and os.path.exists(db.get_1d_flow_path(geo))):
        #     print('Results exist. Skipping...')
        #     continue

        msg = None
        # msg = generate_1d_api(db, geo)
        # sys.exit(0)
        # get parameters
        try:
            params = get_params(db, geo)
            # msg = generate_1d(db, geo)
        except Exception as e:
            msg = 'error: ' + str(e)

        # continue
        if not msg:
            if model_order == 1:
                # run oneDSolver
                start = time.time()
                sv.run_solver_1d(db.get_solve_dir_1d(geo), params['solver_output_file'])
                end = time.time()

                # extract results
                res_dir = db.get_solve_dir_1d(geo)
                params_file = db.get_1d_params(geo)
                results_1d = read_results_1d(res_dir, params_file)

                # save results
                if results_1d['flow']:
                    # save results in dict
                    np.save(db.get_1d_flow_path(geo), results_1d)

                    # remove 1d output files
                    for f in glob.glob(os.path.join(res_dir, geo + 'branch*seg*_*.dat')):
                        os.remove(f)

                    msg = 'success'
                else:
                    msg = 'unconverged'
            elif model_order == 0:
                # parameters for 0d simulation
                zerod_in = os.path.join(db.get_solve_dir_1d(geo), params['solver_output_file'])

                # run 0d simulation
                # try:
                if True:
                    if zerod_cpp:
                        # quick output?
                        with open(zerod_in, 'r') as file:
                            inp = json.load(file)
                            if zerod_cpp_fast:
                                label = True
                            else:
                                label = False
                            inp['simulation_parameters']['output_variable_based'] = label
                            inp['simulation_parameters']['output_last_cycle_only'] = label
                        with open(zerod_in, 'w') as file:
                            json.dump(inp, file, indent=4, sort_keys=True)

                        # run cpp
                        zerod_out = db.get_0d_flow_path(geo).replace('npy', 'csv')

                        start = time.time()
                        err = sv.run_solver_0d(zerod_in, zerod_out)
                        end = time.time()

                        if err:
                            msg = 'cpp error'
                        else:
                            msg = 'success'

                            if not zerod_cpp_fast:
                                # convert to numpy
                                df = read_csv(zerod_out)
                                out = {"flow": {}, "pressure": {}, "distance": {}}

                                # read input file
                                with open(zerod_in, 'r') as file:
                                    inp = json.load(file)

                                # loop branches and segments
                                names = list(sorted(set(df["name"])))
                                for name in names:
                                    # extract ids
                                    br, seg = [int(s) for s in re.findall(r'\d+', name)]

                                    # add 0d results
                                    for field in ['flow', 'pressure']:
                                        if seg == 0:
                                            out[field][br] = [list(df[df.name == name][field + '_in'])]
                                        out[field][br] += [list(df[df.name == name][field + '_out'])]
                                    out["time"] = list(df[df.name == name]["time"])

                                    # add path distance
                                    for vessel in inp['vessels']:
                                        if vessel['vessel_name'] == name:
                                            if seg == 0:
                                                out["distance"][br] = [0]
                                            l_new = out["distance"][br][-1] + vessel['vessel_length']
                                            out["distance"][br] += [l_new]

                                # convert to numpy
                                for field in ['flow', 'pressure', 'distance']:
                                    for br in out[field].keys():
                                        out[field][br] = np.array(out[field][br])
                                out['time'] = np.array(out['time'])

                                # save to file
                                np.save(db.get_0d_flow_path(geo), out)

                    else:
                        ini_steady = True
                        if 'coronary' in db.get_bcs(geo)['bc_type'].values():
                            ini_steady = False

                        start = time.time()
                        run0d.solver.set_up_and_run_0d_simulation(zerod_in,
                                                                  use_steady_soltns_as_ics=ini_steady)
                        end = time.time()

                        # move output file
                        src = os.path.join(db.get_solve_dir_1d(geo), geo + '_0d_branch_results.npy')
                        dst = db.get_0d_flow_path(geo)
                        shutil.move(src, dst)

                        msg = 'success'
                # except Exception as e:
                #     msg = '0d failed: ' + str(e)
        if msg != 'success':
            print('  skipping (1d model creation failed)\n  ' + msg)
        print('\nTime elapsed: ' + '{:.2f}'.format(end - start) + '\n')

        # store errors in file
        if model_order == 1:
            db.add_log_file_1d(geo, msg)
            db.add_log_file_1d(geo + '_time', end - start)
        elif model_order == 0:
            db.add_log_file_0d(geo, msg)
            db.add_log_file_0d(geo + '_time', end - start)


if __name__ == '__main__':
    descr = 'Automatically create, run, and post-process 1d-simulations'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import pdb
import os
import csv
import numpy as np
import json
from collections import defaultdict, OrderedDict

from get_database import Database, Post
from get_sv_project import coronary_sv_to_oned
from common import get_dict
from get_bcs import get_bcs, get_params

import matplotlib.pyplot as plt
import matplotlib.ticker as mtick


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def get_json():
    db = Database()
    geometries = db.get_geometries()
    # geometries = ['0075_1001']
    for geo in geometries:
        print(geo)

        # get all meta-data
        par = db.get_bcs(geo)
        if par is None:
            continue
        par_tcl = db.get_bcs_local(geo)

        # add parameters from tcl that are not in database
        for k, v in par_tcl['params'].items():
            if k not in par['params']:
                # exclude local tcl stuff
                if k not in ['auto_path', 'tk_library', 'tcl_library', 'errorCode', 'tcl_pkgPath']:
                    par['params'][k] = v

        # add inflow
        par['bc']['inflow']['flow'] = db.get_inflow_smooth(geo)

        with open(db.get_json(geo), 'w') as file:
            json.dump(par, file, indent=4, sort_keys=True, cls=NumpyEncoder)


def main():
    db = Database()
    geometries = db.get_geometries()
    geometries_paper = db.get_geometries_select('paper')

    properties = defaultdict(OrderedDict)
    params_read = ['age', 'gender', 'deliverable_category', 'vascular_state', 'treatment', 'other_disease',
                   'sim_physio_state', 'image_data_modality', 'paper_reference',
                   'model_source', 'simulation_source', 'sim_steps_per_cycle']

    # extract simulation parameters
    for geo in geometries:
        print('Running geometry ' + geo)
        bc_def = db.get_bcs(geo)

        for pm in params_read:
            if bc_def['params'] is None or pm not in bc_def['params']:
                val = 'unknown'
            else:
                val = bc_def['params'][pm]
            properties[geo][pm] = val

    # check if simulation is to be published
    for geo in geometries:
        if geo in geometries_paper:
            publish = 'yes'
        else:
            publish = 'no'
        properties[geo]['publish'] = publish

    # get inflow type
    for geo in geometries:
        bc_def = db.get_bcs(geo)
        if bc_def is None:
            inflow = 'none'
        else:
            inflow = bc_def['bc']['inflow']['type']
        properties[geo]['inflow_type'] = inflow

    with open('osmsc2.csv', 'w', newline='') as csvfile:
        reader = csv.writer(csvfile, delimiter=',')

        # write header
        reader.writerow(['model_id'] + list(properties[geometries[0]].keys()))

        # write rows
        for geo in geometries:
            reader.writerow([geo] + [v for v in properties[geo].values()])


def time_constant():
    db = Database('1spb_length')
    post = Post()
    geometries = db.get_geometries()
    # geometries = db.get_geometries_select('paper')

    # get numerical time constants
    res_num = get_dict(db.get_convergence_path())

    fig1, ax1 = plt.subplots(dpi=400, figsize=(10, 3))

    geos = []
    i = 0
    taus = {}
    alphas = {}
    for geo in geometries:
        params = db.get_bcs(geo)

        time, _ = db.get_inflow(geo)

        # collect all time constants
        const = db.get_time_constants(geo)

        # skip geometries without RCR
        if const is None:
            continue

        tau_bc = list(const.values())
        while 0.0 in tau_bc:
            tau_bc.remove(0.0)

        if len(tau_bc) == 0:
            continue

        # skip non-RCR models
        # types = np.unique(list(params['bc_type'].values()))
        # if not len(types) == 1 or not types[0] == 'rcr':
        #     continue

        col = post.colors[params['params']['deliverable_category']]
        if geo in res_num:
            tau_num = res_num[geo]['tau']
            for j in range(len(tau_num['flow'])):
                # ax1.plot(i, tau_num['flow'][j], marker='o', color=col)
                ax1.plot(i, tau_num['pressure'][j], marker='x', color=col)
            taus[geo] = np.mean(tau_num['pressure'])
            alphas[geo] = 1 / (np.exp(1 / taus[geo]) - 1)

            print(geo + '\ttau_num = ' + '{:2.1f}'.format(taus[geo])+ '\talpha_num = ' + '{:2.1f}'.format(alphas[geo]))
        # skip models without 0d solution
        else:
            continue

        # ax1.boxplot(tau_bc, positions=[i])
        ax1.plot([i, i], [np.min(tau_bc), np.max(tau_bc)], '-', color=col)
        ax1.plot(i, np.min(tau_bc), '_', color=col)
        ax1.plot(i, np.max(tau_bc), '_', color=col)

        geos += [geo]
        i += 1

    # plt.yscale('log')
    ax1.xaxis.grid('minor')
    ax1.yaxis.grid(True)
    ax1.set_ylim(0, 10)
    ax1.set_xlim(-1, len(geos))
    plt.xticks(np.arange(len(geos)), geos, rotation='vertical')
    plt.ylabel(r'Time constants [-]')
    # fname = os.path.join(db.fpath_gen, 'time_constants_paper_rcr_with_0d.png')
    # fname = os.path.join(db.fpath_gen, 'time_constants_paper')
    fname = os.path.join(db.fpath_gen, 'time_constants')
    fig1.savefig(fname, bbox_inches='tight')
    pdb.set_trace()

if __name__ == '__main__':
    # main()
    # time_constant()
    get_json()



import pdb
import numpy as np
import matplotlib.pyplot as plt

import sys
import svzerodsolver as run0d


def run_network_util(block_list, deltat, time, bc_type):
    connect_list, wdict = run0d.connections.connect_blocks_by_inblock_list(block_list)
    neq = run0d.connections.compute_neq(block_list, wdict)  # compute number of equations
    for b in block_list:  # run a consistency check
        run0d.connections.check_block_connection(b)
    var_name_list = run0d.connections.assign_global_ids(block_list, wdict)  # assign solution variables with global ID

    y0, ydot0 = run0d.connections.initialize_solution_structures(neq)  # initialize solution structures
    curr_y = y0.copy()
    curr_ydot = ydot0.copy()

    rho = 0.1
    args = {}
    args['Time step'] = deltat
    args['rho'] = rho
    args['Wire dictionary'] = wdict
    args['check_jacobian'] = False

    # create time integration
    t_int = run0d.time_integration.GenAlpha(rho, curr_y)

    ylist = [curr_y.copy()]
    for t in time[1:]:
        args['Solution'] = curr_y
        curr_y, curr_ydot = t_int.step(curr_y, curr_ydot, t, block_list, args, deltat)
        ylist.append(curr_y)

    ###############
    # these corrections are needed based on how i think the above for loop works: the "t" in the above for loop corresponds to t_current (t_n) and the output of gen_alpha_dae_integrator_NR is y_next (y_(n+1)). see https://github.com/StanfordCBCL/0D_LPN_Python_Solver/blob/master/test_nonlin_res.py for proof.
    ylist = ylist[:-1]
    ylist.insert(0, y0.copy())
    ###############

    ylist = np.array(ylist)
    for i in range(len(ylist[0, :])):
        if var_name_list[i] == "P_inflow_" + bc_type:
            inlet_pressure = ylist[:, i]
            return inlet_pressure


def run_rcr(Qfunc, time, p, distal_pressure):
    deltat = time[1] - time[0]

    inflow = run0d.blocks.UnsteadyFlowRef(connecting_block_list=['rcr'],
                                          name='inflow',
                                          Qfunc=Qfunc,
                                          flow_directions=[+1])
    rcr = run0d.blocks.UnsteadyRCRBlockWithDistalPressure(connecting_block_list=['inflow'],
                                                          name='rcr',
                                                          flow_directions=[-1],
                                                          Rp_func=lambda t: p['Rp'],
                                                          C_func=lambda t: p['C'],
                                                          Rd_func=lambda t: p['Rd'],
                                                          Pref_func=lambda t: distal_pressure)
    return run_network_util([inflow, rcr], deltat, time, bc_type="rcr")


def run_coronary(Qfunc, time, p, p_im_time, p_im_pres, cardiac_cycle_period, Pv=0.0):
    deltat = time[1] - time[0]

    Pim_matrix = np.zeros((len(p_im_time), 2))
    Pim_matrix[:, 0] = p_im_time
    Pim_matrix[:, 1] = p_im_pres
    distal_pressure = np.zeros((len(p_im_time), 2))
    distal_pressure[:, 0] = p_im_time
    distal_pressure[:, 1] = np.ones(len(p_im_time)) * Pv  # leaving Pv = 0.0 should be okay

    inflow = run0d.blocks.UnsteadyFlowRef(connecting_block_list=['coronary'], Qfunc=Qfunc, name='inflow',
                                          flow_directions=[+1])

    coronary = run0d.blocks.OpenLoopCoronaryWithDistalPressureBlock(Ra=p['R1'], Ca=p['C1'], Ram=p['R2'], Cim=p['C2'],
                                                                       Rv=p['R3'], Pim=Pim_matrix, Pv=distal_pressure,
                                                                       cardiac_cycle_period=cardiac_cycle_period,
                                                                       connecting_block_list=['inflow'],
                                                                       name='coronary', flow_directions=[-1])

    block_list = [inflow, coronary]

    return run_network_util(block_list, deltat, time, bc_type="coronary")



#!/usr/bin/env python

import pdb
import vtk

import scipy
import numpy as np

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n

from get_database import Database, SimVascular, input_args
from vtk_functions import read_geo, write_geo, collect_arrays, threshold, geo, clean, region_grow, region_grow_simultaneous, ClosestPoints

from collections import defaultdict


def color_clip(poly, name_this, name_other):
    # label array of this name
    color_this = vtk.vtkDoubleArray()
    color_this.SetNumberOfValues(poly.GetNumberOfCells())
    color_this.SetName(name_this)

    # get cell labels from connected points (assume they all have the same label)
    pids = vtk.vtkIdList()
    for i in range(poly.GetNumberOfCells()):
        poly.GetCellPoints(i, pids)
        color_this.SetValue(i, poly.GetPointData().GetArray(name_this).GetValue(pids.GetId(0)))

    # label array of other name (constant -1)
    color_other = vtk.vtkDoubleArray()
    color_other.SetNumberOfValues(poly.GetNumberOfCells())
    color_other.SetName(name_other)
    color_other.Fill(-1)

    poly.GetCellData().AddArray(color_this)
    poly.GetCellData().AddArray(color_other)


def get_bifurcations(cent):
    # read all point arrays
    arr_cent = collect_arrays(cent.GetPointData())

    # all bifurcation ids
    bifurcation_ids = np.unique(arr_cent['BifurcationId']).tolist()
    bifurcation_ids.remove(-1)

    # collect point and cell ids
    pids = vtk.vtkIdList()
    cids = vtk.vtkIdList()

    # collect points surrounding each bifurcation
    bifurcations = {}
    for bf in bifurcation_ids:
        bifurcations[bf] = defaultdict(list)
        for i in range(cent.GetNumberOfPoints()):
            if arr_cent['BifurcationId'][i] == bf:
                cent.GetPointCells(i, cids)
                for j in range(cids.GetNumberOfIds()):
                    cent.GetCellPoints(cids.GetId(j), pids)
                    for k in range(pids.GetNumberOfIds()):
                        br_id = arr_cent['BranchId'][pids.GetId(k)]
                        if br_id != -1:
                            bifurcations[bf]['branches'] += [br_id]
                            bifurcations[bf]['points'] += [pids.GetId(k)]
    return bifurcations


def split_geo(fpath_surf, fpath_cent, fpath_sect, fpath_vol):
    surf = read_geo(fpath_surf).GetOutput()
    cent = read_geo(fpath_cent).GetOutput()
    sect = read_geo(fpath_sect).GetOutput()
    vol = read_geo(fpath_vol).GetOutput()

    arr_surf = collect_arrays(surf.GetPointData())
    arr_cent = collect_arrays(cent.GetPointData())

    bifurcation_ids = np.unique(arr_cent['BifurcationId']).tolist()
    bifurcation_ids.remove(-1)
    branch_ids = np.unique(arr_cent['BranchId']).tolist()
    branch_ids.remove(-1)

    pids = vtk.vtkIdList()
    cids = vtk.vtkIdList()

    # get centerline connectivity
    bifurcations = get_bifurcations(cent)

    branch_ids = arr_surf['BranchId']

    cp = ClosestPoints(surf)

    cut_name = 'distance_bifurcation'

    # distance array used for clipping
    distance = -1 * np.ones(surf.GetNumberOfPoints())

    rings = defaultdict(dict)
    # loop bifurcations
    for bf, bifurcation in bifurcations.items():
        
        # loop attached branches
        # seed_bf = []
        # ring_bf = []
        for i, (p, br) in enumerate(zip(bifurcation['points_global'], bifurcation['branches'])):
            # print(bf, br)

            # pick slice separating bifurcation and branch
            sliced = threshold(sect, p, 'GlobalNodeId').GetOutput()

            # signed distance from slice
            dist = vtk.vtkDistancePolyDataFilter()
            dist.SetInputData(0, surf)
            dist.SetInputData(1, geo(sliced))
            dist.Update()
            dist = v2n(dist.GetOutput().GetPointData().GetArray('Distance'))

            # reverse inlet
            if i > 0:
                dist *= -1

            # pick slice separating bifurcation and branch
            sliced = threshold(sect, p, 'GlobalNodeId').GetOutput()

            # mark bifurcation cuts
            ring = cp.search(v2n(sliced.GetPoints().GetData()))
            # ring_bf += ring

            rings[bf][br] = ring

            # ring_grow_1 = region_grow(surf, ring, distance, 2)
            # ring_grow_2 = region_grow(surf, ring, distance, 3)
            #
            # indicator_1 = np.intersect1d(ring_grow_1, np.where(branch_ids == br))
            # indicator_2 = np.intersect1d(ring_grow_2, np.where(branch_ids == br))

            # bifurcation_id[indicator_1] = 1

            # ring_diff = list(set(indicator_2) - set(indicator_1))
            # seed_bf += np.array(ring_diff)[dist[ring_diff] > 0].tolist()

            # assert np.unique(distance[indicator]).shape[0] == 1, 'overwriting branches'

            # branch_dist[indicator_2] = dist[indicator_2]

        # mark region inside bifurcation
        # bf_ids = region_grow(surf, seed_bf, bifurcation_id)
    bifurcation_id = -1 * np.ones(surf.GetNumberOfPoints())
    branch_dist = np.zeros(surf.GetNumberOfPoints())
    region_grow_simultaneous(surf, rings, bifurcation_id, branch_dist)

    # distance[bf_ids] = 1
    # indicator = branch_dist != -1
    # distance[indicator] = branch_dist[indicator]




    # assemble cutting array
    out = vtk.vtkDoubleArray()
    out.SetNumberOfValues(surf.GetNumberOfPoints())
    out.SetName('branch_dist')
    for i in range(surf.GetNumberOfPoints()):
        out.SetValue(i, branch_dist[i])
    surf.GetPointData().AddArray(out)
    out = vtk.vtkDoubleArray()
    out.SetNumberOfValues(surf.GetNumberOfPoints())
    out.SetName('BifurcationIdTmp')
    for i in range(surf.GetNumberOfPoints()):
        out.SetValue(i, bifurcation_id[i])
    surf.GetPointData().AddArray(out)

    append = vtk.vtkAppendFilter()
    append.AddInputData(surf)
    append.MergePointsOn()
    append.Update()
    return append.GetOutput()

    # assemble cutting array
    out = vtk.vtkDoubleArray()
    out.SetNumberOfValues(surf.GetNumberOfPoints())
    out.SetName(cut_name)
    for i in range(surf.GetNumberOfPoints()):
        out.SetValue(i, distance[i])

    surf.GetPointData().AddArray(out)
    surf.GetPointData().SetActiveScalars(cut_name)

    # clip into branches and bifurcations
    clip = vtk.vtkClipPolyData()
    clip.SetInputData(surf)
    clip.SetValue(0.0)
    clip.GenerateClippedOutputOn()
    clip.GenerateClipScalarsOff()
    clip.Update()

    # color branches and bifurcations
    poly_bf = clip.GetOutput(0)
    poly_br = clip.GetOutput(1)
    color_clip(poly_bf, 'BifurcationId', 'BranchId')
    color_clip(poly_br, 'BranchId', 'BifurcationId')

    # add back together
    append = vtk.vtkAppendFilter()
    append.AddInputData(poly_br)
    append.AddInputData(poly_bf)
    append.MergePointsOn()
    append.Update()

    # remove outdated point arrays
    unstruc_out = append.GetOutput()
    # unstruc_out.GetPointData().RemoveArray(cut_name)
    # unstruc_out.GetPointData().RemoveArray('BranchId')
    # unstruc_out.GetPointData().RemoveArray('BifurcationId')

    unstruc_out.GetCellData().SetActiveScalars('BranchId')

    return unstruc_out


def main(db, geometries):
    for geo in geometries:
        print('Running geometry ' + geo)
        fpath_cent = db.get_centerline_path(geo)
        fpath_surf = db.get_surfaces_grouped_path(geo)
        fpath_sect = db.get_section_path(geo)
        fpath_vol = db.get_volume(geo)

        # try:
        surf_cut = split_geo(fpath_surf, fpath_cent, fpath_sect, fpath_vol)
        # except Exception as e:
        #     print(e)
        #     continue

        write_geo(db.get_surfaces_cut_path(geo), surf_cut)


if __name__ == '__main__':
    descr = 'Split geometry in branches and bifurcation_ids'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import os
import vtk
import pdb

import numpy as np
from collections import defaultdict
from tqdm import tqdm

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n


class Integration:
    """
    Class to perform integration on slices
    """

    def __init__(self, inp):
        try:
            self.integrator = vtk.vtkIntegrateAttributes()
        except AttributeError:
            raise Exception('vtkIntegrateAttributes is currently only supported by pvpython')

        if not inp.GetOutput().GetNumberOfPoints():
            raise Exception('Empty slice')

        self.integrator.SetInputData(inp.GetOutput())
        self.integrator.Update()

    def evaluate(self, res_name):
        """
        Evaluate integral.
        Distinguishes between scalar integration (e.g. pressure) and normal projection (velocity)
        Optionally divides integral by integrated area
        Args:
            field: pressure, velocity, ...
            res_name: name of array

        Returns:
            Scalar integral
        """
        # type of result
        field = res_name.split('_')[0]

        if field == 'velocity':
            int_name = 'normal_' + res_name
        else:
            int_name = res_name

        # evaluate integral
        integral = v2n(self.integrator.GetOutput().GetPointData().GetArray(int_name))[0]

        # choose if integral should be divided by area
        if field == 'velocity':
            return integral
        else:
            return integral / self.area()

    def area(self):
        """
        Evaluate integrated surface area
        Returns:
        Area
        """
        return v2n(self.integrator.GetOutput().GetCellData().GetArray('Area'))[0]


class ClosestPoints:
    """
    Find closest points within a geometry
    """
    def __init__(self, inp):
        if isinstance(inp, str):
            geo = read_geo(inp)
            inp = geo.GetOutput()
        dataset = vtk.vtkPolyData()
        dataset.SetPoints(inp.GetPoints())

        locator = vtk.vtkPointLocator()
        locator.Initialize()
        locator.SetDataSet(dataset)
        locator.BuildLocator()

        self.locator = locator

    def search(self, points, radius=None):
        """
        Get ids of points in geometry closest to input points
        Args:
            points: list of points to be searched
            radius: optional, search radius
        Returns:
            Id list
        """
        ids = []
        for p in points:
            if radius is not None:
                result = vtk.vtkIdList()
                self.locator.FindPointsWithinRadius(radius, p, result)
                ids += [result.GetId(k) for k in range(result.GetNumberOfIds())]
            else:
                ids += [self.locator.FindClosestPoint(p)]
        return ids


def collect_arrays(output):
    res = {}
    for i in range(output.GetNumberOfArrays()):
        name = output.GetArrayName(i)
        data = output.GetArray(i)
        res[name] = v2n(data)
    return res


def get_all_arrays(geo):
    # collect all arrays
    cell_data = collect_arrays(geo.GetCellData())
    point_data = collect_arrays(geo.GetPointData())

    return point_data, cell_data


def read_geo(fname):
    """
    Read geometry from file, chose corresponding vtk reader
    Args:
        fname: vtp surface or vtu volume mesh

    Returns:
        vtk reader, point data, cell data
    """
    _, ext = os.path.splitext(fname)
    if ext == '.vtp':
        reader = vtk.vtkXMLPolyDataReader()
    elif ext == '.vtu':
        reader = vtk.vtkXMLUnstructuredGridReader()
    else:
        raise ValueError('File extension ' + ext + ' unknown.')
    reader.SetFileName(fname)
    reader.Update()

    return reader


def write_geo(fname, input):
    """
    Write geometry to file
    Args:
        fname: file name
    """
    _, ext = os.path.splitext(fname)
    if ext == '.vtp':
        writer = vtk.vtkXMLPolyDataWriter()
    elif ext == '.vtu':
        writer = vtk.vtkXMLUnstructuredGridWriter()
    else:
        raise ValueError('File extension ' + ext + ' unknown.')
    writer.SetFileName(fname)
    writer.SetInputData(input)
    writer.Update()
    writer.Write()


def threshold(inp, t, name):
    """
    Threshold according to cell array
    Args:
        inp: InputConnection
        t: BC_FaceID
        name: name in cell data used for thresholding
    Returns:
        reader, point data
    """
    thresh = vtk.vtkThreshold()
    thresh.SetInputData(inp)
    thresh.SetInputArrayToProcess(0, 0, 0, 1, name)
    thresh.ThresholdBetween(t, t)
    thresh.Update()
    return thresh


def calculator(inp, function, inp_arrays, out_array):
    """
    Function to add vtk calculator
    Args:
        inp: InputConnection
        function: string with function expression
        inp_arrays: list of input point data arrays
        out_array: name of output array
    Returns:
        calc: calculator object
    """
    calc = vtk.vtkArrayCalculator()
    for a in inp_arrays:
        calc.AddVectorArrayName(a)
    calc.SetInputData(inp.GetOutput())
    if hasattr(calc, 'SetAttributeModeToUsePointData'):
        calc.SetAttributeModeToUsePointData()
    else:
        calc.SetAttributeTypeToPointData()
    calc.SetFunction(function)
    calc.SetResultArrayName(out_array)
    calc.Update()
    return calc


def cut_plane(inp, origin, normal):
    """
    Cuts geometry at a plane
    Args:
        inp: InputConnection
        origin: cutting plane origin
        normal: cutting plane normal
    Returns:
        cut: cutter object
    """
    # define cutting plane
    plane = vtk.vtkPlane()
    plane.SetOrigin(origin[0], origin[1], origin[2])
    plane.SetNormal(normal[0], normal[1], normal[2])

    # define cutter
    cut = vtk.vtkCutter()
    cut.SetInputData(inp)
    cut.SetCutFunction(plane)
    cut.Update()
    return cut


def get_points_cells(inp):
    cells = []
    for i in range(inp.GetNumberOfCells()):
        cell_points = []
        for j in range(inp.GetCell(i).GetNumberOfPoints()):
            cell_points += [inp.GetCell(i).GetPointId(j)]
        cells += [cell_points]
    return v2n(inp.GetPoints().GetData()), np.array(cells)


def connectivity(inp, origin):
    """
    If there are more than one unconnected geometries, extract the closest one
    Args:
        inp: InputConnection
        origin: region closest to this point will be extracted
    Returns:
        con: connectivity object
    """
    con = vtk.vtkConnectivityFilter()
    con.SetInputData(inp.GetOutput())
    con.SetExtractionModeToClosestPointRegion()
    con.SetClosestPoint(origin[0], origin[1], origin[2])
    con.Update()
    return con


def connectivity_all(inp):
    """
    Color regions according to connectivity
    Args:
        inp: InputConnection
    Returns:
        con: connectivity object
    """
    con = vtk.vtkConnectivityFilter()
    con.SetInputData(inp)
    con.SetExtractionModeToAllRegions()
    con.ColorRegionsOn()
    con.Update()
    assert con.GetNumberOfExtractedRegions() > 0, 'empty geometry'
    return con


def extract_surface(inp):
    """
    Extract surface from 3D geometry
    Args:
        inp: InputConnection
    Returns:
        extr: vtkExtractSurface object
    """
    extr = vtk.vtkDataSetSurfaceFilter()
    extr.SetInputData(inp)
    extr.Update()
    return extr.GetOutput()


def clean(inp):
    """
    Merge duplicate Points
    """
    cleaner = vtk.vtkCleanPolyData()
    cleaner.SetInputData(inp)
    # cleaner.SetTolerance(1.0e-3)
    cleaner.PointMergingOn()
    cleaner.Update()
    return cleaner.GetOutput()


def scalar_array(length, name, fill):
    """
    Create vtkIdTypeArray array with given name and constant value
    """
    ids = vtk.vtkIdTypeArray()
    ids.SetNumberOfValues(length)
    ids.SetName(name)
    ids.Fill(fill)
    return ids


def add_scalars(inp, name, fill):
    """
    Add constant value array to point and cell data
    """
    inp.GetOutput().GetCellData().AddArray(scalar_array(inp.GetOutput().GetNumberOfCells(), name, fill))
    inp.GetOutput().GetPointData().AddArray(scalar_array(inp.GetOutput().GetNumberOfPoints(), name, fill))


def rename(inp, old, new):
    if inp.GetOutput().GetCellData().HasArray(new):
        inp.GetOutput().GetCellData().RemoveArray(new)
    if inp.GetOutput().GetPointData().HasArray(new):
        inp.GetOutput().GetPointData().RemoveArray(new)
    inp.GetOutput().GetCellData().GetArray(old).SetName(new)
    inp.GetOutput().GetPointData().GetArray(old).SetName(new)


def replace(inp, name, array):
    arr = n2v(array)
    arr.SetName(name)
    inp.GetOutput().GetCellData().RemoveArray(name)
    inp.GetOutput().GetCellData().AddArray(arr)


def geo(inp):
    poly = vtk.vtkGeometryFilter()
    poly.SetInputData(inp)
    poly.Update()
    return poly.GetOutput()


def region_grow(geo, seed_points, seed_ids, n_max=99):
    # initialize output arrays
    array_ids = -1 * np.ones(geo.GetNumberOfPoints(), dtype=int)
    array_rad = np.zeros(geo.GetNumberOfPoints())
    array_dist = -1 * np.ones(geo.GetNumberOfPoints(), dtype=int)
    array_ids[seed_points] = seed_ids

    # initialize ids
    cids_all = set()
    pids_all = set(seed_points.tolist())
    pids_new = set(seed_points.tolist())

    # get points
    pts = v2n(geo.GetPoints().GetData())

    # loop until region stops growing or reaches maximum number of iterations
    i = 0
    while len(pids_new) > 0 and i < n_max:
        # update
        pids_old = pids_new

        # print progress
        print_str = 'Iteration ' + str(i)
        print_str += '\tNew points ' + str(len(pids_old)) + '     '
        print_str += '\tTotal points ' + str(len(pids_all))
        print(print_str)

        # grow region one step
        pids_new = grow(geo, array_ids, pids_old, pids_all, cids_all)

        # convert to array
        pids_old_arr = list(pids_old)

        # create point locator with old wave front
        points = vtk.vtkPoints()
        points.Initialize()
        for i_old in pids_old:
            points.InsertNextPoint(geo.GetPoint(i_old))

        dataset = vtk.vtkPolyData()
        dataset.SetPoints(points)

        locator = vtk.vtkPointLocator()
        locator.Initialize()
        locator.SetDataSet(dataset)
        locator.BuildLocator()

        # find closest point in new wave front
        for i_new in pids_new:
            i_old = pids_old_arr[locator.FindClosestPoint(geo.GetPoint(i_new))]
            array_ids[i_new] = array_ids[i_old]
            array_rad[i_new] = array_rad[i_old] + np.linalg.norm(pts[i_new] - pts[i_old])
            array_dist[i_new] = i

        # count grow iterations
        i += 1

    return array_ids, array_dist + 1, array_rad


def grow(geo, array, pids_in, pids_all, cids_all):
    # ids of propagating wave-front
    pids_out = set()

    # loop all points in wave-front
    for pi_old in pids_in:
        cids = vtk.vtkIdList()
        geo.GetPointCells(pi_old, cids)

        # get all connected cells in wave-front
        for j in range(cids.GetNumberOfIds()):
            # get cell id
            ci = cids.GetId(j)

            # skip cells that are already in region
            if ci in cids_all:
                continue
            else:
                cids_all.add(ci)
            
            pids = vtk.vtkIdList()
            geo.GetCellPoints(ci, pids)

            # loop all points in cell
            for k in range(pids.GetNumberOfIds()):
                # get point id
                pi_new = pids.GetId(k)

                # add point only if it's new and doesn't fullfill stopping criterion
                if array[pi_new] == -1 and pi_new not in pids_in:
                    pids_out.add(pi_new)
                    pids_all.add(pi_new)

    return pids_out


def cell_connectivity(geo):
    """
    Extract the point connectivity from vtk and return a dictionary that can be used in meshio
    """
    vtk_to_meshio = {3: 'line', 5: 'triangle', 10: 'tetra'}

    cells = defaultdict(list)
    for i in range(geo.GetNumberOfCells()):
        cell_type_vtk = geo.GetCellType(i)
        if cell_type_vtk in vtk_to_meshio:
            cell_type = vtk_to_meshio[cell_type_vtk]
        else:
            raise ValueError('vtkCellType ' + str(cell_type_vtk) + ' not supported')

        points = geo.GetCell(i).GetPointIds()
        point_ids = []
        for j in range(points.GetNumberOfIds()):
            point_ids += [points.GetId(j)]
        cells[cell_type] += [point_ids]

    for t, c in cells.items():
        cells[t] = np.array(c)

    return cells



#!/usr/bin/env python

try:
    import sv
except ImportError:
    raise ImportError('Run with sv --python -- this_script.py')

import argparse
import glob
import os


def main(p):
    sv.PathLegacyIO.Convert(file_name=p.path_in, output_dir=p.path_out)
    for seg in glob.glob(os.path.join(p.seg_in, '*')):
        if '.tcl' not in seg:
            sv.SegmentationLegacyIO.Convert(file_name=seg, path_file=p.path_in, output_dir=p.seg_out)


if __name__ == '__main__':
    # parse input arguments
    parser = argparse.ArgumentParser(description='convert legacy path in SimVascular')
    parser.add_argument('path_in', help='input path folder')
    parser.add_argument('path_out', help='output path folder')
    parser.add_argument('seg_in', help='input segmentation folder')
    parser.add_argument('seg_out', help='output segmentation folder')

    # run script
    main(parser.parse_args())



#!/usr/bin/env python

import contextlib
import csv
import glob
import io
import os
import pdb
import re
import shutil
import sys
import argparse
import subprocess

import numpy as np

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n

from get_database import Database, SimVascular, Post, input_args
from simulation_io import get_dict
from vtk_functions import read_geo, write_geo, collect_arrays

import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
from matplotlib.offsetbox import OffsetImage, AnnotationBbox


def print_error(db, geometries):
    # get post-processing constants
    post = Post()

    folder = db.get_statistics_dir()

    # get simulation errors
    m_rom = '0d'
    if m_rom == '0d':
        res_all = get_dict(db.get_0d_3d_comparison())
    if m_rom == '1d':
        res_all = get_dict(db.get_1d_3d_comparison())

    # use only geometries in selection
    res = {}
    for k, v in res_all.items():
        if k in geometries:
            res[k] = v

    # make plots
    # plot_error_spatial(db, geometries)
    # plot_error_centerline(db, res, folder)
    plot_err_bar(res, folder)
    plot_scatter(db, res, 'pts')
    plot_scatter(db, res, 'img')


def plot_error_centerline(db, res, folder):
    # get post-processing constants
    post = Post()

    metric0 = ['avg', 'max']#, 'sys', 'dia'

    for k in res:
        # read centerline
        cent = read_geo(db.get_centerline_path(k)).GetOutput()
        arrays = collect_arrays(cent.GetPointData())
        branches = arrays['BranchId']

        for f in post.fields:
            for m0 in metric0:
                # write error to centerline
                out = np.zeros(cent.GetNumberOfPoints())
                for br, err in res[k][f]['spatial'][m0].items():
                    out[branches == br] = err

                # add array to centerline
                array = n2v(out)
                array.SetName(f + '_' + m0)
                cent.GetPointData().AddArray(array)

        # write to file
        write_geo(os.path.join(folder, k + '.vtp'), cent)


def plot_err_bar(res, folder):
    # get post-processing constants
    post = Post()

    domain = ['cap', 'int']
    metric0 = ['avg', 'max']#, 'sys', 'dia'

    fig1, ax1 = plt.subplots(dpi=300, figsize=(12, 6))
    for f in post.fields:
        for d in domain:
            for m0 in metric0:
                labels = []
                values = []
                for k in res:
                    labels += [k]
                    values += [res[k][f][d][m0]['all']]

                labels = np.array(labels)
                values = np.array(values)
                xtick = np.arange(len(values))

                order = np.argsort(values)
                plot_bar(fig1, ax1, xtick, values, labels, order, m0, f, d, folder, 'sorted')

                order = np.argsort(labels)
                plot_bar(fig1, ax1, xtick, values, labels, order, m0, f, d, folder, 'aplhabetical')
    plt.close(fig1)


def plot_bar(fig1, ax1, xtick, values, labels, order, m0, f, d, folder, name):
    plt.cla()
    plt.yscale('log')
    ax1.bar(xtick, values[order])
    ax1.yaxis.grid(True)
    ax1.yaxis.set_major_formatter(mtick.PercentFormatter(xmax=1, decimals=1))
    ax1.set_ylim(0.001, 1)
    plt.xticks(xtick, labels[order], rotation='vertical')
    plt.ylabel(m0 + ' ' + f + ' error at ' + d + ' [1]')
    fname = os.path.join(folder, 'error_' + name + '_' + f + '_' + d + '_' + m0 + '.png')
    fig1.savefig(fname, bbox_inches='tight')


def plot_scatter(db, res, mode):
    fsize = 70
    fig1, ax1 = plt.subplots(dpi=100, figsize=(60, 30))
    plt.rcParams.update({'font.size': fsize})
    plt.rcParams['axes.linewidth'] = 2

    combinations = [['flow', 'pressure']]#, ['area', 'flow'], ['area', 'pressure']]
    domain = {'cap': 'at caps', 'int': 'in branches'}
    metric0 = ['avg', 'max']#, 'sys', 'dia'

    for c in combinations:
        fx = c[0]
        fy = c[1]
        for d in domain:
            for m0 in metric0:
                plt.cla()
                for geo, err in res.items():
                    x = err[fx][d][m0]['all']
                    y = err[fy][d][m0]['all']
                    if mode == 'img':
                        ab = AnnotationBbox(OffsetImage(plt.imread(db.get_png(geo))), (x, y), frameon=False)
                        ax1.scatter(x, y, c='k')
                        ax1.add_artist(ab)
                    elif mode == 'pts':
                        ax1.plot(x, y, 'o')
                        ax1.annotate(geo, (x, y))

                plt.xlabel(fx.capitalize() + ' relative ' + m0 + '. error ' + domain[d], fontsize=fsize)
                plt.ylabel(fy.capitalize() + ' relative ' + m0 + '. error ' + domain[d], fontsize=fsize)
                plt.xscale('log')
                plt.yscale('log')
                plt.grid(b=True, which='major', color='k', linestyle='-', linewidth=2)
                plt.grid(b=True, which='minor', color='0.5', linestyle='-', linewidth=0.5)
                ax1.xaxis.set_major_formatter(mtick.PercentFormatter(xmax=1, decimals=0))
                ax1.yaxis.set_major_formatter(mtick.PercentFormatter(xmax=1, decimals=0))
                ax1.set_xlim(0.006, 0.3)
                ax1.set_ylim(0.003, 0.3)

                fname = 'error_correlation_' + fx + '_' + fy + '_' + d + '_' + m0 + '_' + mode + '.png'
                fpath = os.path.join(db.get_statistics_dir(), fname)
                fig1.savefig(fpath, bbox_inches='tight')
    plt.close(fig1)


def plot_error_spatial(db, geometries):
    # set global plot options
    fig, ax = plt.subplots(2, 2, figsize=(14, 8), dpi=200)
    plt.rcParams['axes.linewidth'] = 2

    # read errors
    f_path = db.get_3d_3d_comparison()
    if not os.path.exists(f_path):
        return
    err = get_dict(f_path)

    # select geometries
    legend = [geo for geo in geometries if geo in err]

    for i, f in enumerate(['pressure', 'velocity']):
        for j, c in enumerate(['avg', 'max']):
            # plot location
            pos = (j, i)

            # plot data points
            lengths = []
            for geo in legend:
                e = err[geo][f][c]
                x = np.arange(1, len(e) + 1)
                lengths += [len(e)]
                ax[pos].plot(x, e, 'o-')


            # set plot options
            ax[pos].grid(True)
            ax[pos].set_xlabel('Cycle [1]')
            ax[pos].set_ylabel(f.capitalize() + ' relative ' + c + ' error [1]')
            ax[pos].ticklabel_format(axis='y', style='sci', scilimits=(0, 0))
            ax[pos].legend(legend)
            # ax[pos].set_yscale('log')
            ax[pos].set_xticks(np.arange(1, np.max(lengths) + 1))
    fname = '3d_3d_comparison.png'
    fpath = os.path.join(db.get_statistics_dir(), fname)
    fig.savefig(fpath, bbox_inches='tight')
    plt.close(fig)



def print_statistics(db, geometries):
    res_all = get_dict(db.get_log_file_1d())

    # use only geometries in selection
    res = {}
    for k, v in res_all.items():
        if k in geometries:
            res[k] = v

    # sort errors
    success = '1D simulation\nsuccessful'
    for k, v in res.items():
        if '3d geometry has multiple inlets' in v:
            res[k] = 'Multiple inlets'
        if 'Inlet group id is not 0 or number of centerlines is not equal to the number of outlets' in v:
            res[k] = 'Centerline extraction\nfailed'
        if 'float division by zero' in v:
            res[k] = '3D geometry is corrupted'
        if 'boundary conditions not implemented (coronary)' in v:
            res[k] = 'Coronary\nboundary conditions'
        if 'The number of BC values' in v:
            res[k] = 'Missing boundary conditions'
        if 'unconverged' in v:
            res[k] = '1D simulation\nunconverged'
        if 'bifurcation with less than 2 outflows detected' in v:
            res[k] = 'Bifurcation is at outlet'
        if 'KeyError(None,)' in v:
            res[k] = 'Bifurcation is at inlet'
        if 'Centerline consist of more than one region' in v:
            res[k] = 'Centerline consists of >1 piece'
        if 'success' in v:
            res[k] = success
        if 'loop' in v:
            res[k] = '3D geometry contains a loop'

    errors = np.array([k for k in res.values()])

    # count errors
    num_errors = {}
    for err in np.unique(errors):
        num_errors[err] = {}
        num_errors[err]['n'] = np.sum(errors == err)
        num_errors[err]['geos'] = [k for k, v in res.items() if v == err]

    # remove no bcs
    if '3d results do not exist' in num_errors:
        del num_errors['3d results do not exist']

    for err, geos in num_errors.items():
        print(err)
        print(geos['geos'])

    for err in num_errors.keys():
        g_string = err + '\n'
        for g in num_errors[err]['geos']:
            g_string += g + '\n'
        num_errors[err]['g_string'] = g_string[:-1]

    # make a montage for every error with the geometries
    montage = num_errors.copy()
    montage['all'] = {'geos': geometries}
    for e in montage.keys():
        err = e.replace(' ', '_')
        err = err.replace('\n', '_')

        g_string = ['/usr/bin/montage']
        for g in montage[e]['geos']:
            extensions = ['_sim.png', '_sim.jpg', '_model.jpg', '_vol.png']
            for ext in extensions:
                src = os.path.join(db.fpath_png, 'OSMSC' + g + ext)
                if os.path.exists(src):
                    break
            g_string += [src]
        g_string += [os.path.join(db.get_statistics_dir(), 'models_' + err.lower()) + '.png']
        subprocess.Popen(g_string)

    # print statistics
    num_sim = len(res)
    print('number of simulations: ' + repr(num_sim))

    geo_id = [geo[:6] for geo in geometries]
    print('number of unique geometries: ' + repr(np.unique(geo_id).shape[0]))

    geo_pat = [geo[:4] for geo in geometries]
    print('number of unique patients: ' + repr(np.unique(geo_pat).shape[0]))

    # Pie chart, where the slices will be ordered and plotted counter-clockwise:
    labels_error = list(num_errors.keys())
    labels = [num_errors[v]['g_string'] for v in labels_error]
    sizes = [num_errors[v]['n'] for v in labels_error]
    # assert np.sum(np.array(sizes)) == num_sim, 'wrong number of errors'

    explode = [0] * len(labels_error)
    explode[labels_error.index(success)] = 0.1
    explode = tuple(explode)

    fig1, ax1 = plt.subplots(dpi=300, figsize=(4, 4))
    ax1.axis('equal')
    # autopct = '%1.1f%%'
    autopct = lambda p: '{:.0f}'.format(p * np.sum(sizes) / 100)

    labels_pie = ax1.pie(sizes, explode=explode, labels=labels_error, autopct=autopct, startangle=90)
    #textprops={'weight': 'bold', labeldistance=1)

    # for label in labels_pie[1]:
    #     label.set_horizontalalignment('center')

    # plt.tight_layout()

    fig1.savefig(os.path.join(db.get_statistics_dir(), 'statistics.png'), bbox_inches='tight')

    # plt.cla()
    # ax1.pie(sizes, explode=explode, labels=labels, autopct=lambda p: '{:.0f}'.format(p * num_sim / 100), startangle=90)
    # fig1.savefig(os.path.join(db.get_statistics_dir(), 'statistics_geometries.png'))

    plt.close()


def main(db, geometries, params):
    # print_statistics(db, geometries)
    print_error(db, geometries)

if __name__ == '__main__':
    descr = 'Automatically create, run, and post-process 1d-simulations'
    d, g, p = input_args(descr)
    main(d, g, p)
